"""Tests of interaction of matrix with other parts of numpy.

Note that tests with MaskedArray and linalg are done in separate files.
"""
import pytest

import textwrap
import warnings

import numpy as np
from numpy.testing import (assert_, assert_equal, assert_raises,
                           assert_raises_regex, assert_array_equal,
                           assert_almost_equal, assert_array_almost_equal)


def test_fancy_indexing():
    # The matrix class messes with the shape. While this is always
    # weird (getitem is not used, it does not have setitem nor knows
    # about fancy indexing), this tests gh-3110
    # 2018-04-29: moved here from core.tests.test_index.
    m = np.matrix([[1, 2], [3, 4]])

    assert_(isinstance(m[[0, 1, 0], :], np.matrix))

    # gh-3110. Note the transpose currently because matrices do *not*
    # support dimension fixing for fancy indexing correctly.
    x = np.asmatrix(np.arange(50).reshape(5, 10))
    assert_equal(x[:2, np.array(-1)], x[:2, -1].T)


def test_polynomial_mapdomain():
    # test that polynomial preserved matrix subtype.
    # 2018-04-29: moved here from polynomial.tests.polyutils.
    dom1 = [0, 4]
    dom2 = [1, 3]
    x = np.matrix([dom1, dom1])
    res = np.polynomial.polyutils.mapdomain(x, dom1, dom2)
    assert_(isinstance(res, np.matrix))


def test_sort_matrix_none():
    # 2018-04-29: moved here from core.tests.test_multiarray
    a = np.matrix([[2, 1, 0]])
    actual = np.sort(a, axis=None)
    expected = np.matrix([[0, 1, 2]])
    assert_equal(actual, expected)
    assert_(type(expected) is np.matrix)


def test_partition_matrix_none():
    # gh-4301
    # 2018-04-29: moved here from core.tests.test_multiarray
    a = np.matrix([[2, 1, 0]])
    actual = np.partition(a, 1, axis=None)
    expected = np.matrix([[0, 1, 2]])
    assert_equal(actual, expected)
    assert_(type(expected) is np.matrix)


def test_dot_scalar_and_matrix_of_objects():
    # Ticket #2469
    # 2018-04-29: moved here from core.tests.test_multiarray
    arr = np.matrix([1, 2], dtype=object)
    desired = np.matrix([[3, 6]], dtype=object)
    assert_equal(np.dot(arr, 3), desired)
    assert_equal(np.dot(3, arr), desired)


def test_inner_scalar_and_matrix():
    # 2018-04-29: moved here from core.tests.test_multiarray
    for dt in np.typecodes['AllInteger'] + np.typecodes['AllFloat'] + '?':
        sca = np.array(3, dtype=dt)[()]
        arr = np.matrix([[1, 2], [3, 4]], dtype=dt)
        desired = np.matrix([[3, 6], [9, 12]], dtype=dt)
        assert_equal(np.inner(arr, sca), desired)
        assert_equal(np.inner(sca, arr), desired)


def test_inner_scalar_and_matrix_of_objects():
    # Ticket #4482
    # 2018-04-29: moved here from core.tests.test_multiarray
    arr = np.matrix([1, 2], dtype=object)
    desired = np.matrix([[3, 6]], dtype=object)
    assert_equal(np.inner(arr, 3), desired)
    assert_equal(np.inner(3, arr), desired)


def test_iter_allocate_output_subtype():
    # Make sure that the subtype with priority wins
    # 2018-04-29: moved here from core.tests.test_nditer, given the
    # matrix specific shape test.

    # matrix vs ndarray
    a = np.matrix([[1, 2], [3, 4]])
    b = np.arange(4).reshape(2, 2).T
    i = np.nditer([a, b, None], [],
                  [['readonly'], ['readonly'], ['writeonly', 'allocate']])
    assert_(type(i.operands[2]) is np.matrix)
    assert_(type(i.operands[2]) is not np.ndarray)
    assert_equal(i.operands[2].shape, (2, 2))

    # matrix always wants things to be 2D
    b = np.arange(4).reshape(1, 2, 2)
    assert_raises(RuntimeError, np.nditer, [a, b, None], [],
                  [['readonly'], ['readonly'], ['writeonly', 'allocate']])
    # but if subtypes are disabled, the result can still work
    i = np.nditer([a, b, None], [],
                  [['readonly'], ['readonly'],
                   ['writeonly', 'allocate', 'no_subtype']])
    assert_(type(i.operands[2]) is np.ndarray)
    assert_(type(i.operands[2]) is not np.matrix)
    assert_equal(i.operands[2].shape, (1, 2, 2))


def like_function():
    # 2018-04-29: moved here from core.tests.test_numeric
    a = np.matrix([[1, 2], [3, 4]])
    for like_function in np.zeros_like, np.ones_like, np.empty_like:
        b = like_function(a)
        assert_(type(b) is np.matrix)

        c = like_function(a, subok=False)
        assert_(type(c) is not np.matrix)


def test_array_astype():
    # 2018-04-29: copied here from core.tests.test_api
    # subok=True passes through a matrix
    a = np.matrix([[0, 1, 2], [3, 4, 5]], dtype='f4')
    b = a.astype('f4', subok=True, copy=False)
    assert_(a is b)

    # subok=True is default, and creates a subtype on a cast
    b = a.astype('i4', copy=False)
    assert_equal(a, b)
    assert_equal(type(b), np.matrix)

    # subok=False never returns a matrix
    b = a.astype('f4', subok=False, copy=False)
    assert_equal(a, b)
    assert_(not (a is b))
    assert_(type(b) is not np.matrix)


def test_stack():
    # 2018-04-29: copied here from core.tests.test_shape_base
    # check np.matrix cannot be stacked
    m = np.matrix([[1, 2], [3, 4]])
    assert_raises_regex(ValueError, 'shape too large to be a matrix',
                        np.stack, [m, m])


def test_object_scalar_multiply():
    # Tickets #2469 and #4482
    # 2018-04-29: moved here from core.tests.test_ufunc
    arr = np.matrix([1, 2], dtype=object)
    desired = np.matrix([[3, 6]], dtype=object)
    assert_equal(np.multiply(arr, 3), desired)
    assert_equal(np.multiply(3, arr), desired)


def test_nanfunctions_matrices():
    # Check that it works and that type and
    # shape are preserved
    # 2018-04-29: moved here from core.tests.test_nanfunctions
    mat = np.matrix(np.eye(3))
    for f in [np.nanmin, np.nanmax]:
        res = f(mat, axis=0)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (1, 3))
        res = f(mat, axis=1)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (3, 1))
        res = f(mat)
        assert_(np.isscalar(res))
    # check that rows of nan are dealt with for subclasses (#4628)
    mat[1] = np.nan
    for f in [np.nanmin, np.nanmax]:
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            res = f(mat, axis=0)
            assert_(isinstance(res, np.matrix))
            assert_(not np.any(np.isnan(res)))
            assert_(len(w) == 0)

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            res = f(mat, axis=1)
            assert_(isinstance(res, np.matrix))
            assert_(np.isnan(res[1, 0]) and not np.isnan(res[0, 0])
                    and not np.isnan(res[2, 0]))
            assert_(len(w) == 1, 'no warning raised')
            assert_(issubclass(w[0].category, RuntimeWarning))

        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            res = f(mat)
            assert_(np.isscalar(res))
            assert_(res != np.nan)
            assert_(len(w) == 0)


def test_nanfunctions_matrices_general():
    # Check that it works and that type and
    # shape are preserved
    # 2018-04-29: moved here from core.tests.test_nanfunctions
    mat = np.matrix(np.eye(3))
    for f in (np.nanargmin, np.nanargmax, np.nansum, np.nanprod,
              np.nanmean, np.nanvar, np.nanstd):
        res = f(mat, axis=0)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (1, 3))
        res = f(mat, axis=1)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (3, 1))
        res = f(mat)
        assert_(np.isscalar(res))

    for f in np.nancumsum, np.nancumprod:
        res = f(mat, axis=0)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (3, 3))
        res = f(mat, axis=1)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (3, 3))
        res = f(mat)
        assert_(isinstance(res, np.matrix))
        assert_(res.shape == (1, 3*3))


def test_average_matrix():
    # 2018-04-29: moved here from core.tests.test_function_base.
    y = np.matrix(np.random.rand(5, 5))
    assert_array_equal(y.mean(0), np.average(y, 0))

    a = np.matrix([[1, 2], [3, 4]])
    w = np.matrix([[1, 2], [3, 4]])

    r = np.average(a, axis=0, weights=w)
    assert_equal(type(r), np.matrix)
    assert_equal(r, [[2.5, 10.0/3]])


def test_dot_matrix():
    # Test to make sure matrices give the same answer as ndarrays
    # 2018-04-29: moved here from core.tests.test_function_base.
    x = np.linspace(0, 5)
    y = np.linspace(-5, 0)
    mx = np.matrix(x)
    my = np.matrix(y)
    r = np.dot(x, y)
    mr = np.dot(mx, my.T)
    assert_almost_equal(mr, r)


def test_ediff1d_matrix():
    # 2018-04-29: moved here from core.tests.test_arraysetops.
    assert(isinstance(np.ediff1d(np.matrix(1)), np.matrix))
    assert(isinstance(np.ediff1d(np.matrix(1), to_begin=1), np.matrix))


def test_apply_along_axis_matrix():
    # this test is particularly malicious because matrix
    # refuses to become 1d
    # 2018-04-29: moved here from core.tests.test_shape_base.
    def double(row):
        return row * 2

    m = np.matrix([[0, 1], [2, 3]])
    expected = np.matrix([[0, 2], [4, 6]])

    result = np.apply_along_axis(double, 0, m)
    assert_(isinstance(result, np.matrix))
    assert_array_equal(result, expected)

    result = np.apply_along_axis(double, 1, m)
    assert_(isinstance(result, np.matrix))
    assert_array_equal(result, expected)


def test_kron_matrix():
    # 2018-04-29: moved here from core.tests.test_shape_base.
    a = np.ones([2, 2])
    m = np.asmatrix(a)
    assert_equal(type(np.kron(a, a)), np.ndarray)
    assert_equal(type(np.kron(m, m)), np.matrix)
    assert_equal(type(np.kron(a, m)), np.matrix)
    assert_equal(type(np.kron(m, a)), np.matrix)


class TestConcatenatorMatrix:
    # 2018-04-29: moved here from core.tests.test_index_tricks.
    def test_matrix(self):
        a = [1, 2]
        b = [3, 4]

        ab_r = np.r_['r', a, b]
        ab_c = np.r_['c', a, b]

        assert_equal(type(ab_r), np.matrix)
        assert_equal(type(ab_c), np.matrix)

        assert_equal(np.array(ab_r), [[1, 2, 3, 4]])
        assert_equal(np.array(ab_c), [[1], [2], [3], [4]])

        assert_raises(ValueError, lambda: np.r_['rc', a, b])

    def test_matrix_scalar(self):
        r = np.r_['r', [1, 2], 3]
        assert_equal(type(r), np.matrix)
        assert_equal(np.array(r), [[1, 2, 3]])

    def test_matrix_builder(self):
        a = np.array([1])
        b = np.array([2])
        c = np.array([3])
        d = np.array([4])
        actual = np.r_['a, b; c, d']
        expected = np.bmat([[a, b], [c, d]])

        assert_equal(actual, expected)
        assert_equal(type(actual), type(expected))


def test_array_equal_error_message_matrix():
    # 2018-04-29: moved here from testing.tests.test_utils.
    with pytest.raises(AssertionError) as exc_info:
        assert_equal(np.array([1, 2]), np.matrix([1, 2]))
    msg = str(exc_info.value)
    msg_reference = textwrap.dedent("""\

    Arrays are not equal

    (shapes (2,), (1, 2) mismatch)
     ACTUAL: array([1, 2])
     DESIRED: matrix([[1, 2]])""")
    assert_equal(msg, msg_reference)


def test_array_almost_equal_matrix():
    # Matrix slicing keeps things 2-D, while array does not necessarily.
    # See gh-8452.
    # 2018-04-29: moved here from testing.tests.test_utils.
    m1 = np.matrix([[1., 2.]])
    m2 = np.matrix([[1., np.nan]])
    m3 = np.matrix([[1., -np.inf]])
    m4 = np.matrix([[np.nan, np.inf]])
    m5 = np.matrix([[1., 2.], [np.nan, np.inf]])
    for assert_func in assert_array_almost_equal, assert_almost_equal:
        for m in m1, m2, m3, m4, m5:
            assert_func(m, m)
            a = np.array(m)
            assert_func(a, m)
            assert_func(m, a)



from numpy._core.strings import (
    equal as equal,
    not_equal as not_equal,
    greater_equal as greater_equal,
    less_equal as less_equal,
    greater as greater,
    less as less,
    add as add,
    multiply as multiply,
    mod as mod,
    isalpha as isalpha,
    isalnum as isalnum,
    isdigit as isdigit,
    isspace as isspace,
    isnumeric as isnumeric,
    isdecimal as isdecimal,
    islower as islower,
    isupper as isupper,
    istitle as istitle,
    str_len as str_len,
    find as find,
    rfind as rfind,
    index as index,
    rindex as rindex,
    count as count,
    startswith as startswith,
    endswith as endswith,
    decode as decode,
    encode as encode,
    expandtabs as expandtabs,
    center as center,
    ljust as ljust,
    rjust as rjust,
    lstrip as lstrip,
    rstrip as rstrip,
    strip as strip,
    zfill as zfill,
    upper as upper,
    lower as lower,
    swapcase as swapcase,
    capitalize as capitalize,
    title as title,
    replace as replace,
    join as join,
    split as split,
    rsplit as rsplit,
    splitlines as splitlines,
    partition as partition,
    rpartition as rpartition,
    translate as translate,
)

__all__: list[str]



#cython: language_level=3

"""
Functions in this module give python-space wrappers for cython functions
exposed in numpy/__init__.pxd, so they can be tested in test_cython.py
"""
cimport numpy as cnp
cnp.import_array()


def is_td64(obj):
    return cnp.is_timedelta64_object(obj)


def is_dt64(obj):
    return cnp.is_datetime64_object(obj)


def get_dt64_value(obj):
    return cnp.get_datetime64_value(obj)


def get_td64_value(obj):
    return cnp.get_timedelta64_value(obj)


def get_dt64_unit(obj):
    return cnp.get_datetime64_unit(obj)


def is_integer(obj):
    return isinstance(obj, (cnp.integer, int))


def get_datetime_iso_8601_strlen():
    return cnp.get_datetime_iso_8601_strlen(0, cnp.NPY_FR_ns)


def convert_datetime64_to_datetimestruct():
    cdef:
        cnp.npy_datetimestruct dts
        cnp.PyArray_DatetimeMetaData meta
        cnp.int64_t value = 1647374515260292
        # i.e. (time.time() * 10**6) at 2022-03-15 20:01:55.260292 UTC

    meta.base = cnp.NPY_FR_us
    meta.num = 1
    cnp.convert_datetime64_to_datetimestruct(&meta, value, &dts)
    return dts


def make_iso_8601_datetime(dt: "datetime"):
    cdef:
        cnp.npy_datetimestruct dts
        char result[36]  # 36 corresponds to NPY_FR_s passed below
        int local = 0
        int utc = 0
        int tzoffset = 0

    dts.year = dt.year
    dts.month = dt.month
    dts.day = dt.day
    dts.hour = dt.hour
    dts.min = dt.minute
    dts.sec = dt.second
    dts.us = dt.microsecond
    dts.ps = dts.as = 0

    cnp.make_iso_8601_datetime(
        &dts,
        result,
        sizeof(result),
        local,
        utc,
        cnp.NPY_FR_s,
        tzoffset,
        cnp.NPY_NO_CASTING,
    )
    return result


cdef cnp.broadcast multiiter_from_broadcast_obj(object bcast):
    cdef dict iter_map = {
        1: cnp.PyArray_MultiIterNew1,
        2: cnp.PyArray_MultiIterNew2,
        3: cnp.PyArray_MultiIterNew3,
        4: cnp.PyArray_MultiIterNew4,
        5: cnp.PyArray_MultiIterNew5,
    }
    arrays = [x.base for x in bcast.iters]
    cdef cnp.broadcast result = iter_map[len(arrays)](*arrays)
    return result


def get_multiiter_size(bcast: "broadcast"):
    cdef cnp.broadcast multi = multiiter_from_broadcast_obj(bcast)
    return multi.size


def get_multiiter_number_of_dims(bcast: "broadcast"):
    cdef cnp.broadcast multi = multiiter_from_broadcast_obj(bcast)
    return multi.nd


def get_multiiter_current_index(bcast: "broadcast"):
    cdef cnp.broadcast multi = multiiter_from_broadcast_obj(bcast)
    return multi.index


def get_multiiter_num_of_iterators(bcast: "broadcast"):
    cdef cnp.broadcast multi = multiiter_from_broadcast_obj(bcast)
    return multi.numiter


def get_multiiter_shape(bcast: "broadcast"):
    cdef cnp.broadcast multi = multiiter_from_broadcast_obj(bcast)
    return tuple([multi.dimensions[i] for i in range(bcast.nd)])


def get_multiiter_iters(bcast: "broadcast"):
    cdef cnp.broadcast multi = multiiter_from_broadcast_obj(bcast)
    return tuple([<cnp.flatiter>multi.iters[i] for i in range(bcast.numiter)])


def get_default_integer():
    if cnp.NPY_DEFAULT_INT == cnp.NPY_LONG:
        return cnp.dtype("long")
    if cnp.NPY_DEFAULT_INT == cnp.NPY_INTP:
        return cnp.dtype("intp")
    return None

def get_ravel_axis():
    return cnp.NPY_RAVEL_AXIS


def conv_intp(cnp.intp_t val):
    return val


def get_dtype_flags(cnp.dtype dtype):
    return dtype.flags


cdef cnp.NpyIter* npyiter_from_nditer_obj(object it):
    """A function to create a NpyIter struct from a nditer object.

    This function is only meant for testing purposes and only extracts the
    necessary info from nditer to test the functionality of NpyIter methods
    """
    cdef:
        cnp.NpyIter* cit
        cnp.PyArray_Descr* op_dtypes[3]
        cnp.npy_uint32 op_flags[3]
        cnp.PyArrayObject* ops[3]
        cnp.npy_uint32 flags = 0

    if it.has_index:
        flags |= cnp.NPY_ITER_C_INDEX
    if it.has_delayed_bufalloc:
        flags |= cnp.NPY_ITER_BUFFERED | cnp.NPY_ITER_DELAY_BUFALLOC
    if it.has_multi_index:
        flags |= cnp.NPY_ITER_MULTI_INDEX

    # one of READWRITE, READONLY and WRTIEONLY at the minimum must be specified for op_flags
    for i in range(it.nop):
        op_flags[i] = cnp.NPY_ITER_READONLY

    for i in range(it.nop):
        op_dtypes[i] = cnp.PyArray_DESCR(it.operands[i])
        ops[i] = <cnp.PyArrayObject*>it.operands[i]

    cit = cnp.NpyIter_MultiNew(it.nop, &ops[0], flags, cnp.NPY_KEEPORDER,
                               cnp.NPY_NO_CASTING, &op_flags[0],
                               <cnp.PyArray_Descr**>NULL)
    return cit


def get_npyiter_size(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = cnp.NpyIter_GetIterSize(cit)
    cnp.NpyIter_Deallocate(cit)
    return result


def get_npyiter_ndim(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = cnp.NpyIter_GetNDim(cit)
    cnp.NpyIter_Deallocate(cit)
    return result


def get_npyiter_nop(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = cnp.NpyIter_GetNOp(cit)
    cnp.NpyIter_Deallocate(cit)
    return result


def get_npyiter_operands(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    try:
        arr = cnp.NpyIter_GetOperandArray(cit)
        return tuple([<cnp.ndarray>arr[i] for i in range(it.nop)])
    finally:
        cnp.NpyIter_Deallocate(cit)


def get_npyiter_itviews(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = tuple([cnp.NpyIter_GetIterView(cit, i) for i in range(it.nop)])
    cnp.NpyIter_Deallocate(cit)
    return result


def get_npyiter_dtypes(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    try:
        arr = cnp.NpyIter_GetDescrArray(cit)
        return tuple([<cnp.dtype>arr[i] for i in range(it.nop)])
    finally:
        cnp.NpyIter_Deallocate(cit)


def npyiter_has_delayed_bufalloc(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = cnp.NpyIter_HasDelayedBufAlloc(cit)
    cnp.NpyIter_Deallocate(cit)
    return result


def npyiter_has_index(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = cnp.NpyIter_HasIndex(cit)
    cnp.NpyIter_Deallocate(cit)
    return result


def npyiter_has_multi_index(it: "nditer"):
    cdef cnp.NpyIter* cit = npyiter_from_nditer_obj(it)
    result = cnp.NpyIter_HasMultiIndex(cit)
    cnp.NpyIter_Deallocate(cit)
    return result


def npyiter_has_finished(it: "nditer"):
    cdef cnp.NpyIter* cit
    try:
        cit = npyiter_from_nditer_obj(it)
        cnp.NpyIter_GotoIterIndex(cit, it.index)
        return not (cnp.NpyIter_GetIterIndex(cit) < cnp.NpyIter_GetIterSize(cit))
    finally:
        cnp.NpyIter_Deallocate(cit)

def compile_fillwithbyte():
    # Regression test for gh-25878, mostly checks it compiles.
    cdef cnp.npy_intp dims[2]
    dims = (1, 2)
    pos = cnp.PyArray_ZEROS(2, dims, cnp.NPY_UINT8, 0)
    cnp.PyArray_FILLWBYTE(pos, 1)
    return pos

def inc2_cfloat_struct(cnp.ndarray[cnp.cfloat_t] arr):
    # This works since we compile in C mode, it will fail in cpp mode
    arr[1].real += 1
    arr[1].imag += 1
    # This works in both modes
    arr[1].real = arr[1].real + 1
    arr[1].imag = arr[1].imag + 1



# intersection of two circles with same r
from numpy import (linspace as _np_linspace, array as _np_array, abs as _np_abs, dot as _np_dot, roll as _np_roll)
from math import sin as _math_sin, cos as _math_cos, pi as _math_pi
from matplotlib import pyplot as plt
from aabpl.utils.general import angles_to_origin, angle_to, pt_is_left_of_vector
from aabpl.utils.rotations import transform_cell_pattern, transform_coord, transform_cell
from aabpl.utils.intersections import circle_line_segment_intersection, line_intersection, intersections_pts_arc_to_circle
from shapely.geometry import Polygon, LineString, Point
from geopandas import GeoSeries


# for each circle remember the meaning of the check (contains / overlaps)
# for each intersection point save what circle it comes from
# for each circle get intersection points with home cell
# for each circle-intersection point check whether its in triangle 1


class Vertex(object):
    """
    2D coordinate
    """

    def __init__(self,x,y,all_vtx):
        """
        create Vertex at coordinate
        """
        self.x = x
        self.y = y
        self.xy = (x,y)
        # self.outgoing_edges = []
        # self.incoming_edges = []
        self.regions = []
        self.all_vtx = all_vtx
        if not self.xy in all_vtx:
            all_vtx[self.xy] = self
        else:
            pass
            # self = all_vtx[self.xy]
    
    @staticmethod
    def clear(all_vtx):
        """
        clear all_vtx
        """
        all_vtx.clear()
    #

    @staticmethod
    def plot_many(vertices:dict=None, ax=None):
        """
        plot vertices at list of coordinates. if no list provided instead plots all stored in all_vtx
        """
        if ax is None:
          fig, ax = plt.subplots()

        vertices = list(vertices.values())
        if len(vertices):
            ax.scatter(x=[vtx.x for vtx in vertices], y=[vtx.y for vtx in vertices], marker='x', color='black')
        #
        return ax

    #
    
    def __repr__(self):
        props_not_to_print = ['all_vtx', 'outgoing_edges', 'incoming_edges', 'regions']
        return "xy:"+str(tuple([round(a,5) for a in self.xy]))
        # return str({key: val if type(val) != float else round(val,5) for key, val in self.__dict__.items() if key not in props_not_to_print})
    #

    def delete(self):
        """
        Remove vertex from all_vtx
        """
        self.all_vtx.pop(self.xy, None)
    #
#


class Edge(object):
    
    def __init__(
            self,
            vtx1:Vertex,
            vtx2:Vertex,
            all_edges:dict,
            contains:tuple=None,
            overlaps:tuple=None,
            enforce_int:bool=True
            ):
        """
        vertices are order counter clockwise
        enforce_int ensures that contains and overlaps are tuples of int() not numpy.int
        """
        self.vtx1 = vtx1
        self.vtx2 = vtx2
        self.coords = (
            vtx1.xy if (vtx1 is not None) else (None, None), 
            vtx2.xy if (vtx2 is not None) else (None, None),
            )
        self.regions = []
        # vtx1.outgoing_edges.append(self)
        # vtx2.incoming_edges.append(self)
        self.all_edges = all_edges
        self.all_vtx = vtx1.all_vtx
        
        if not self.coords in all_edges:
            all_edges[self.coords] = self

        if not contains is None:
            self.contains  = tuple([int(z) for z in contains]) if enforce_int else contains
        if not overlaps is None:
            self.overlaps  = tuple([int(z) for z in overlaps]) if enforce_int else overlaps
    #

    @staticmethod
    def clear(all_edges):
        all_edges.clear()
    #

    def __repr__(self):
        props_not_to_print = ['all_edges', 'all_vtx', 'regions','coords', 'angle_min', 'angle_max']
        return str({key: val if type(val) != float else round(val,5) for key, val in self.__dict__.items() if key not in props_not_to_print})
    #
    
    def delete(self):
        self.all_edges.pop(self.coords, None)
    #

    @staticmethod
    def plot_many(
        edges:dict,
        ax=None,
        color_dict={'Arc':'blue', 'LineSegment':'green','Circle':'orange', 'other':'red'},
        **kwargs
        ):
        if ax is None:
          fig, ax = plt.subplots()
        #
        arc_polys = []
        for coords,edge in edges.items():
            edge.plot_single(ax=ax, color_dict=color_dict, **kwargs)
        if len(arc_polys) > 0:
            GeoSeries(arc_polys).plot(cmap='viridis', alpha=0.01, edgecolor='black', ax=ax)
        return ax
        #
    #

    def plot_single(self, ax=None, color_dict={'Arc':'blue', 'LineSegment':'green','Circle':'orange', 'other':'red'}, **kwargs):
        if ax is None:
          fig, ax = plt.subplots()
        #
        arc_polys=[]
        if self.type == 'Arc':
            arc_polys.append(Polygon([self.center, self.coords[0], self.coords[1]]))

            # ax.plot([0, self.center[0]], [0, self.center[1]], marker='.', color="black")
            rotation = angle_to((0,0), self.center)
            
            # ax.plot(
            #     [self.center[0], self.center[0]+self.r*_math_cos((rotation+self.angle_min)/360*2*_math_pi)],
            #     [self.center[1], self.center[1]+self.r*_math_sin((rotation+self.angle_min)/360*2*_math_pi)],
            #     marker='.', color='red',alpha=0.3,)
            # ax.plot(
            #     [self.center[0], self.center[0]+self.r*_math_cos((rotation+self.angle_max)/360*2*_math_pi)],
            #     [self.center[1], self.center[1]+self.r*_math_sin((rotation+self.angle_max)/360*2*_math_pi)],
            #     marker='.', color='green',alpha=0.3,)
            ax.plot(
                [self.center[0], self.center[0]+self.r*_math_cos((rotation+self.angle_vtx1)/360*2*_math_pi)],
                [self.center[1], self.center[1]+self.r*_math_sin((rotation+self.angle_vtx1)/360*2*_math_pi)],
                marker='.', color='red',alpha=0.3,)
            ax.plot(
                [self.center[0], self.center[0]+self.r*_math_cos((rotation+self.angle_vtx2)/360*2*_math_pi)],
                [self.center[1], self.center[1]+self.r*_math_sin((rotation+self.angle_vtx2)/360*2*_math_pi)],
                marker='.', color='green',alpha=0.3,)
            x_coords = [x for x,y in self.coords]
            y_coords = [y for x,y in self.coords]
            ax.plot(x_coords, y_coords, marker='o', color=color_dict[self.type],alpha=0.3, **kwargs)
        else:
            GeoSeries([LineString((self.vtx1.xy, self.vtx2.xy))]).plot(ax=ax, color='pink')
        return arc_polys
    #
#

class LineSegment(Edge):
    """LineSegment between two 2D coordinates"""
    def __init__(
            self,
            vtx1:Vertex,
            vtx2:Vertex,
            all_edges:dict,
            contains:tuple=None,
            overlaps:tuple=None
        ):
        super().__init__(vtx1, vtx2, all_edges, contains=contains, overlaps=overlaps)
        self.type = 'LineSegment'
    #

    def intersection(self,edge):
        """
        
        """
        if edge.type == 'LineSegment':
            return line_intersection(self.coords, edge.coords)
        
        return circle_line_segment_intersection(
            circle_center=edge.center,
            circle_radius=edge.r,
            pt1=self.vtx1.xy,
            pt2=self.vtx2.xy,
            full_line=False,
        )
        
    def split(self, new_vtx):
        """Split at point if point is not start or end point of line segment"""
        if new_vtx.xy in self.coords:
            return [self]
        self.all_edges.pop(self.coords, None)
        line_kwargs = {'all_edges': self.all_edges, **{key: getattr(self, key) for key in ['contains', 'overlaps'] if hasattr(self, key)}}
        return [LineSegment(vtx1=self.vtx1, vtx2=new_vtx, **line_kwargs), LineSegment(vtx1=new_vtx, vtx2=self.vtx2, **line_kwargs)]
    #

    def calc_min_dist_to_pt(self, pt):
        """
        returns smallest distance between edge (a finite line segement) and pt
        """
        px = self.vtx2.x-self.vtx1.x
        py = self.vtx2.y-self.vtx1.y
        u =  ((pt[0] - self.vtx1.x) * px + (pt[1] - self.vtx1.y) * py) / float(px**2 + py**2)
        if u > 1:
            u = 1
        elif u < 0:
            u = 0
        return (((self.vtx1.x + u * px) - pt[0])**2 + ((self.vtx1.y + u * py) - pt[1])**2)**.5
    #

    def calc_max_dist_to_pt(self, pt):
        """
        returns largest distance between point and edge (=finite line segment)
        """
        return max([
            ((self.vtx1.x-pt[0])**2+(self.vtx1.y-pt[1])**2)**.5,
            ((self.vtx2.x-pt[0])**2+(self.vtx2.y-pt[1])**2)**.5
        ])
    #

    def transform_to_trgl(self, i:int):
        """
        
        """
        if i == 1: return self 
        
        new_vtx1 = Vertex(*transform_coord(self.vtx1.xy, i), all_vtx=self.all_vtx)
        new_vtx2 = Vertex(*transform_coord(self.vtx2.xy, i), all_vtx=self.all_vtx)
        
        edge_kwargs = {key: tuple(transform_cell(cell=_np_array(getattr(self, key)), i=i)) for key in ['contains', 'overlaps'] if hasattr(self, key)}
        if i%2 == 1: return LineSegment(vtx1=new_vtx1, vtx2=new_vtx2, all_edges=self.all_edges, **edge_kwargs)
        # flip order for regions 2,4,6,8
        return LineSegment(vtx1=new_vtx2, vtx2=new_vtx1, all_edges=self.all_edges, **edge_kwargs)
#


class Arc(Edge):
    """Arc around center (limited by two points on circle if supplied) """
    def __init__(
            self,
            center:tuple,
            r:float,
            all_edges:dict,
            vtx1:Vertex=None,
            vtx2:Vertex=None,
            clockwise:bool=False,
            contains:tuple=None,
            overlaps:tuple=None,
            ):
        super().__init__(vtx1, vtx2, all_edges, contains=contains, overlaps=overlaps)
        self.type = "Arc"
        self.center = center
        self.r = r
        self.vtx1 = vtx1
        self.vtx2 = vtx2

        self.angle_vtx1, self.angle_vtx2  = angles_to_origin((vtx1.xy, vtx2.xy), center) if not None in [vtx1,vtx2] else (0,360)
        self.angle_min, self.angle_max = sorted((self.angle_vtx1, self.angle_vtx2))
        self.clockwise = clockwise # To-Do.
    #

    def intersection(self, edge):
        """
        
        """
  
        if edge.type == 'Arc':
            return intersections_pts_arc_to_circle(
                circle_center=edge.center,
                arc_center=self.center,
                arc_angle_min=self.angle_min,
                arc_angle_max=self.angle_max,
                r=self.r,
            )
        elif edge.type == 'Circle':
            return intersections_pts_arc_to_circle(
                circle_center=edge.center,
                arc_center=self.center,
                arc_angle_min=self.angle_min,
                arc_angle_max=self.angle_max,
                r=self.r,
            )
        
        return circle_line_segment_intersection(
            circle_center=self.center,
            circle_radius=self.r,
            pt1=edge.vtx1.xy,
            pt2=edge.vtx2.xy,
            full_line=False,
        )
    
    def split(self, new_vtx):
        """Split at point if point is not start or end point of arc segment"""
        if new_vtx.xy in self.coords:
            return [self]
        self.all_edges.pop(self.coords, None)
        arc_kwargs = {
            'center': self.center,
            'r': self.r,
            'all_edges': self.all_edges,
            'clockwise': self.clockwise,
            **{key: getattr(self, key) for key in ['contains', 'overlaps'] if hasattr(self, key)}
            }

        return [
            Arc(vtx1=self.vtx1, vtx2=new_vtx, **arc_kwargs), 
            Arc(vtx1=new_vtx, vtx2=self.vtx2, **arc_kwargs)
            ]
    #
    
    def calc_min_dist_to_pt(self, pt):
        """
        returns smallest distance between edge (an arc, not full circle) and pt
        """
        angle_pt = angles_to_origin([pt], self.center)[0]
        if self.angle_min < angle_pt and angle_pt < self.angle_max: # TODO double check if this condition is sufficient
            return min([
            ((self.vtx1.x-pt[0])**2+(self.vtx1.y-pt[1])**2)**.5,
            ((self.vtx2.x-pt[0])**2+(self.vtx2.y-pt[1])**2)**.5,
            ((pt[0]-self.center[0])**2 + (pt[1]-self.center[1])**2)**.5 - self.r,
        ])
        return min([
            ((self.vtx1.x-pt[0])**2+(self.vtx1.y-pt[1])**2)**.5,
            ((self.vtx2.x-pt[0])**2+(self.vtx2.y-pt[1])**2)**.5
        ])
    #

    def calc_max_dist_to_pt(self, pt):
        """
        returns largest distance between point and edge (=arc, not full circle)
        """
        angle_pt = angles_to_origin([pt], self.center)[0]
        if self.angle_min < angle_pt and angle_pt < self.angle_max: # TODO double check if this condition is sufficient
            return max([
            ((self.vtx1.x-pt[0])**2+(self.vtx1.y-pt[1])**2)**.5,
            ((self.vtx2.x-pt[0])**2+(self.vtx2.y-pt[1])**2)**.5,
            ((pt[0]-self.center[0])**2 + (pt[1]-self.center[1])**2)**.5 - self.r,
        ])
        return max([
            ((self.vtx1.x-pt[0])**2+(self.vtx1.y-pt[1])**2)**.5,
            ((self.vtx2.x-pt[0])**2+(self.vtx2.y-pt[1])**2)**.5
        ])
    #
    
    def transform_to_trgl(self, i:int):
        """
        
        """
        if i == 1: return self 
        new_center = transform_coord(self.center, i)
        new_vtx1 = Vertex(*transform_coord(self.vtx1.xy, i), all_vtx=self.all_vtx)
        new_vtx2 = Vertex(*transform_coord(self.vtx2.xy, i), all_vtx=self.all_vtx)
        # flip order for regions 2,4,6,8
        edge_kwargs = {key: tuple(transform_cell(cell=_np_array(getattr(self, key)), i=i)) for key in ['contains', 'overlaps'] if hasattr(self, key)}
        if i%2 == 1: return Arc(
            new_center,
            r=self.r,
            all_edges=self.all_edges,
            vtx1=new_vtx1,
            vtx2=new_vtx2,
            clockwise=self.clockwise,
            **edge_kwargs
            )
        return Arc(
            new_center,
            r=self.r,
            all_edges=self.all_edges,
            vtx1=new_vtx2,
            vtx2=new_vtx1,
            clockwise=not self.clockwise,
            **edge_kwargs
            )
#

class Circle(object):
    """TODO potentially add as subclass of Edge?"""
    def __init__(self, center, r):
        self.type = "Circle"
        self.center = center
        self.r = r
        self.angle_min, self.angle_max = (0,360)
    #
#

class OffsetRegion(object):
    """OffsetRegion bounded by line segments and arcs"""

    def __init__(self, edges, checks, all_regions:dict, trgl_nr:int=1):
        self.id = -1
        self.edges = edges
        self.vertices = []
        self.trgl_nr = trgl_nr
        xs, ys = [], []
        for edge in edges:
            edge.regions.append(self)
            for vtx in (edge.vtx1, edge.vtx2):
                if vtx not in self.vertices:
                    self.vertices.append(vtx)
                    vtx.regions.append(self)
            # check if most extreme x/y values are not at the vertices but between
            if edge.type == 'Arc':
                # To-Do use clockwise information!
                if edge.angle_vtx1 < 90 and 90 < edge.angle_vtx2:
                    ys.append(edge.center[1] + edge.r)
                if edge.angle_vtx1 < 180 and 180 < edge.angle_vtx2:
                    xs.append(edge.center[0] - edge.r)
                if edge.angle_vtx1 < 270 and 270 < edge.angle_vtx2:
                    ys.append(edge.center[1] - edge.r)
                if edge.angle_vtx1 > edge.angle_vtx2:
                    xs.append(edge.center[0] + edge.r)
        # if len(xs):
        #     print("xs",xs,min([vtx.x for vtx in self.vertices]),max([vtx.x for vtx in self.vertices]))
        # if len(ys):
        #     print(angles_to_origin([(1,1),(1,0),(0,-1)],(0,0)))
        #     print("ys",ys, min([vtx.y for vtx in self.vertices]),max([vtx.y for vtx in self.vertices]))
        # To-Do: For edges that have an arc it might be that a point between the vertices is 
        # the max/min x/y value. Probably only the case  
        xs = [vtx.x for vtx in self.vertices] #+ xs
        ys = [vtx.y for vtx in self.vertices] #+ ys
        self.xmin = min(xs)
        self.xmax = max(xs)
        self.ymin = min(ys)
        self.ymax = max(ys)
        self.checks = checks
        self.coords = tuple([edge.coords for edge in edges])
        self.all_regions = all_regions
        self.all_edges = edges[-1].all_edges
        self.all_vtx = edges[-1].vtx1.all_vtx

        if not self.coords in all_regions:
            all_regions[self.coords] = self
    #
    
    @staticmethod
    def delete_all(all_regions):
        all_regions.clear()
    #

    def __repr__(self):
        props_not_to_print = ['all_regions', 'all_edges', 'all_vtx', 'plot_coords']
        rnd = 5
        return str({key: {val.type:[(round(x,rnd),round(y,rnd)) for x,y in [val.vtx1,val.vtx2]]} if val in ['edges'] else (val if type(val) != float else round(val,rnd)) for key, val in self.__dict__.items() if key not in props_not_to_print})
    #

    def delete(self):
        self.all_regions.pop(self.coords, None)
    #

    def get_coords(self):
        coords = []
        for edge_start_coord, edge_end_coord in self.coords:
            if edge_start_coord not in coords:
                coords.append(edge_start_coord)
            if edge_start_coord not in coords:
                coords.append(edge_start_coord)
        self.vertex_coords = coords
        return coords
    #

    def get_plot_coords(self, arc_steps_per_degree:float=50):
        coords = []
        for (edge_start_coord, edge_end_coord), edge in zip(self.coords, self.edges):
            if edge_start_coord not in coords:
                coords.append(edge_start_coord)
            if edge.type == 'Arc':
                total_angle = abs(edge.angle_max - edge.angle_min)
                n_steps = -int(-(total_angle * arc_steps_per_degree))
                cx, cy = edge.center
                r = edge.r
                rotation = angle_to((0,0), edge.center)
                for angle_step in _np_linspace(edge.angle_min, edge.angle_max, n_steps):
                    x = cx + r * _math_cos((rotation+angle_step)/360*2*_math_pi)
                    y = cy + r * _math_sin((rotation+angle_step)/360*2*_math_pi)
                    coords.append((float(x),float(y)))
            if edge_end_coord not in coords:
                coords.append(edge_start_coord)
        self.plot_coords = coords
        return coords
    #

    def calc_area(self, arc_steps_per_degree:float=100):
        xs = []
        ys = []
        for ((x,y), edge_end_coord), edge in zip(self.coords, self.edges):
            xs.append(x)  
            ys.append(y)  
            if edge.type == 'Arc':
                total_angle = abs(edge.angle_max - edge.angle_min)
                n_steps = -int(-(total_angle * arc_steps_per_degree))
                cx, cy = edge.center
                r = edge.r
                rotation = angle_to((0,0), edge.center)
                for angle_step in _np_linspace(edge.angle_min, edge.angle_max, n_steps):
                    xs.append(cx + r * _math_cos((rotation+angle_step)/360*2*_math_pi))
                    ys.append(cy + r * _math_sin((rotation+angle_step)/360*2*_math_pi))
        xs =_np_array(xs)
        ys =_np_array(ys)
        self.area = 0.5*_np_abs(_np_dot(xs,_np_roll(ys,1))-_np_dot(ys,_np_roll(xs,1))) 
        return self.area

    @staticmethod
    def plot_many(
        regions:list, 
        plot_edges:bool=False, plot_vertices:bool=True, add_idxs:bool=True, x_lim:tuple=(-0.02,0.52), y_lim:tuple=(-0.02,0.52),
        arc_steps_per_degree:float=50,
        ax=None, cmap='tab20', alpha=0.3, title:str="", **kwargs):

        if ax is None:
            fig, ax = plt.subplots()
        
        if len(regions) > 0:
            color_keys = ['c', 'color']
            if not any([key in color_keys for key in kwargs]):
                kwargs.update({'cmap': cmap})
            
            
            try:
                # GeoSeries([Polygon(reg.get_plot_coords()) for reg in regions]).plot(ax=ax, alpha=alpha, **{'edgecolor':'black', **kwargs})
                GeoSeries([Polygon(reg.get_plot_coords(arc_steps_per_degree=arc_steps_per_degree)) for reg in regions]).plot(ax=ax, alpha=alpha, **kwargs)
            except:
                raise ValueError("kwargs", kwargs, "polys", [Polygon(reg.get_plot_coords(arc_steps_per_degree=arc_steps_per_degree)) for reg in regions],"ax", type(ax), ax, "alpha", alpha)
            if add_idxs:
                for i, region in enumerate(regions):
                    coords = region.get_coords()
                    ax.annotate(
                        text=str(i if not hasattr(region, 'id') else region.id),
                        xy=(sum([x for x,y in coords])/len(coords), sum([y for x,y in coords])/len(coords)),
                        horizontalalignment='center')
                
        if plot_edges:
            Edge.plot_many(edges=regions[-1].all_edges, ax=ax,markersize=0.2)
        if plot_vertices:
            # for reg in regions.values():
            #     Vertex.plot_many(all_vtx=[vtx.xy for vtx in reg.vertices], ax=ax)
            Vertex.plot_many(vertices=regions[-1].all_vtx, ax=ax)
        ax.set_xlim(x_lim)
        ax.set_ylim(y_lim)
        ax.set_title(title or str(len(regions)) + " regions.")
        return ax

    #

    def plot_single(self,
                    ax=None,
                    plot_edges:bool=True,
                    plot_vertices:bool=False,
                    add_idx_edges:bool=False,
                    x_lim:tuple=(-0.02,0.52),
                    y_lim:tuple=(-0.02,0.52),
                    color='green',
                    edgecolor='black',
                    alpha=0.70, 
                    **kwargs):
        if ax is None:
            fig, ax = plt.subplots()
        kwargs = {'color': color, 'alpha': alpha, 'edgecolor': edgecolor,  **kwargs }
        GeoSeries([Polygon(self.get_plot_coords())]).plot(ax=ax, **kwargs)
        if plot_edges:
            for edge in self.edges:
                edge.plot_single(ax=ax)
        if plot_vertices:
            Vertex.plot_many(vertices=[vtx.xy for vtx in self.vertices], ax=ax)
        if add_idx_edges:
            for i, edge in enumerate(self.edges):
                ax.annotate(text=str(i), xy=edge.vtx1.xy) 
        if not x_lim is None:
            ax.set_xlim(x_lim)
        if not y_lim is None:
            ax.set_ylim(y_lim)
    #


    def get_split_pts(self, intersection_edge):
        pts_to_split_at, vertices_intersected = [], []

        for i, edge in enumerate(self.edges):
            itx = edge.intersection(intersection_edge)
            # print(i,itx, edge)
            # check if itx is in edge
            if len(itx) == 0:
                continue 
            #

            if len(itx) == 1:
                pt = itx[0]
                # TODO check if vertex already exists: then its clear
                # new_vtx = Vertex(*itx[0])
                if not pt in vertices_intersected:
                    pts_to_split_at.append({'i': i, 'pt': pt, 'edge': edge, 'touches': pt in edge.coords, 'ids':[i]})# [edge1, itx[0], edge2], 
                else:
                    for pt_to_split_at in pts_to_split_at:
                        if pt_to_split_at['pt'] == pt:
                            pt_to_split_at['pt']['ids'].append(i)
                            print("pt_to_split_at['pt']['ids']",pt_to_split_at['pt']['ids'])
                            break
                if not pt in edge.coords:
                    vertices_intersected.append(pt)
                #
                continue
            #

            if len(itx) == 2:
                pts_to_split_at.append({'i': i, 'pts': itx, 'edge': edge, 'touches': pt in edge.coords})
                print("two intersections:", itx)
                continue
            #

            raise ValueError("TOO MANY (",len(itx),") INTERSECTIONS!", itx)
        #
        return pts_to_split_at
    #


    
    def add_check_result_no_intersection(self, intersection_edge, check):
        """
        
        """

        # mean_x = sum([c[0][0] for c in self.coords]) / len(self.coords)
        # mean_y = sum([c[0][1] for c in self.coords]) / len(self.coords)
        vtx_not_on_intersection_edge = next(
            (coord for coord in [startcoord for startcoord, endcoord in self.coords] if (
            hasattr(intersection_edge, 'vtx1') and coord != intersection_edge.vtx1.xy and coord != intersection_edge.vtx2.xy)),
            self.coords[0][0])
        if intersection_edge.type == 'LineSegment':
            # TODO this must store better on which side of the line implies check success
            result = pt_is_left_of_vector(*vtx_not_on_intersection_edge, *intersection_edge.vtx1.xy, *intersection_edge.vtx2.xy)
        else:
            x,y=vtx_not_on_intersection_edge
            # TODO as polygon is not necessarily convex centroid may lay outide, s.t this check may produce wrong results.  
            # BETTER APPROACH. Chose a point that is not on the current intersection edge. 
            # print('r', intersection_edge.r, ((mean_x-intersection_edge.center[0])**2 + (mean_y-intersection_edge.center[1])**2)**.5)
            result = intersection_edge.r**2 > ((x - intersection_edge.center[0])**2 + (y - intersection_edge.center[1])**2)
        self.checks.append({**check, 'result': result})
    #

    def split_with_circle(self, intersection_edge, check, plot_split:bool=False):
        """
        check if intersected
        split edges
        add vertices
        split region 
        TODO rework this logic. Too convoluted!
        TODO THIS FUNCTION IS WAY TOO LONG
        """
        pts_to_split_at = self.get_split_pts(intersection_edge=intersection_edge)
        vertices_intersected, split_pts = [], []

        for split_at in pts_to_split_at:
            if 'pt' in split_at:
                if not split_at['pt'] in vertices_intersected:
                    split_pts.append(split_at)
                vertices_intersected.append(split_at['pt'])
            else:
                split_pts.append(split_at)
        
        pts_to_split_at=split_pts
        # print("Length after filter", len(pts_to_split_at), pts_to_split_at)
        if len(pts_to_split_at) == 0:
            self.add_check_result_no_intersection(intersection_edge=intersection_edge, check=check)
            return 'green' if self.checks[-1]['result'] else 'red'
        
        if len(pts_to_split_at) > 2:
            raise ValueError("TODO IMPLEMENT MULTIPLE INTERSECTIONS", pts_to_split_at, intersection_edge.center)
        
        line_kwargs = {
            'all_edges': self.all_edges, 
            **{key: getattr(intersection_edge, key) for key in ['contains', 'overlaps'] if hasattr(intersection_edge, key)}
            }
        arc_kwargs = {'center': intersection_edge.center, 'r': intersection_edge.r, **line_kwargs}

        if len(pts_to_split_at) == 1:
            if not 'pts' in pts_to_split_at[0]:
                self.add_check_result_no_intersection(intersection_edge=intersection_edge, check=check)
                return 'green' if self.checks[-1]['result'] else 'red'
            
            if len(pts_to_split_at[0]['pts']) != 2:
                raise ValueError("Unexpected number of itx", pts_to_split_at)
            print("OPTION A")
            start_pt, end_pt = pts_to_split_at[0]['pts']
            old_edge = pts_to_split_at[0]['edge']
            old_edge.delete()
            pos = pts_to_split_at[0]['i']
            start_vtx_existed = start_pt not in self.all_vtx
            end_vtx_existed = end_pt not in self.all_vtx

            start_new_vtx = Vertex(*start_pt, self.all_vtx) if start_vtx_existed else self.all_vtx[start_pt]
            end_new_vtx = Vertex(*end_pt, self.all_vtx) if end_vtx_existed else self.all_vtx[end_pt] 
            
            # if  type(intersection_edge) == Arc:
            if type(intersection_edge) != LineSegment:
                new_edge_start =               [Arc(vtx1=old_edge.vtx1, vtx2=start_new_vtx, clockwise=old_edge.clockwise, **arc_kwargs)] if start_vtx_existed else []
                new_edge_end =                 [Arc(vtx1=end_new_vtx,   vtx2=old_edge.vtx2, clockwise=old_edge.clockwise, **arc_kwargs)] if end_vtx_existed else []
                new_edge_middle =              [Arc(vtx1=start_new_vtx, vtx2=end_new_vtx, clockwise=old_edge.clockwise,   **arc_kwargs)]
                new_edge_intersection =         Arc(vtx1=start_new_vtx, vtx2=end_new_vtx, clockwise=True or False,   **arc_kwargs)
                new_edge_intersection_reverse = Arc(vtx1=end_new_vtx,   vtx2=start_new_vtx, clockwise=False or True, **arc_kwargs)
            else:
                new_edge_start =               [LineSegment(vtx1=old_edge.vtx1, vtx2=start_new_vtx, **line_kwargs)] if start_vtx_existed else []
                new_edge_end =                 [LineSegment(vtx1=end_new_vtx,   vtx2=old_edge.vtx2, **line_kwargs)] if end_vtx_existed else []
                new_edge_middle =              [LineSegment(vtx1=end_new_vtx,   vtx2=old_edge.vtx2, **line_kwargs)]
                new_edge_intersection =         LineSegment(vtx1=start_new_vtx, vtx2=end_new_vtx,   **line_kwargs)
                new_edge_intersection_reverse = LineSegment(vtx1=end_new_vtx,   vtx2=start_new_vtx, **line_kwargs)
            
            # TODO MAKE THIS HANDLE LINESEGEMENTS ASWELL
            first_region_is_within_radius = pt_is_left_of_vector(*new_edge_intersection.center, *new_edge_intersection.vtx1.xy, *new_edge_intersection.vtx2.xy)
            # region without intersection_edge
            reg_without_itx = OffsetRegion(
                edges = self.edges[:pos] + new_edge_start + [new_edge_intersection] + new_edge_end + self.edges[pos+1:], 
                checks = self.checks+[{**check, 'result': first_region_is_within_radius}], all_regions = self.all_regions
            ) 
            # new intersection region 
            reg_itx = OffsetRegion(
                edges = new_edge_middle + [new_edge_intersection_reverse], 
                checks = self.checks+[{**check, 'result': not first_region_is_within_radius}], all_regions = self.all_regions
            )
            if plot_split:
                fig, ax = plt.subplots()
                self.plot_single(ax=ax, color='#caa', plot_edges=False)
                reg_without_itx.plot_single(ax=ax, color='green', alpha=0.5, hatch='/', plot_edges=False)
                reg_itx.plot_single(ax=ax, color='red', alpha=0.5, hatch='\\', plot_edges=False)
                GeoSeries([Point(intersection_edge.center).buffer(intersection_edge.r, 40)]).plot(ax=ax, edgecolor='black', facecolor='None')
                ax.set_title("1")
            self.delete()

            return 'yellow'
        
        if len(pts_to_split_at) == 2:
            start, end = pts_to_split_at
            if any(['pts' in d for d in pts_to_split_at]):
                raise ValueError("multiple double intersections", pts_to_split_at)
            start_pt, end_pt = start['pt'], end['pt'] 
            
            if start['touches'] and end['touches'] and (
                (start_pt, end_pt) in self.coords or (end_pt, start_pt) in self.coords or start_pt==end_pt
            ):
                indexes = []
                if (start_pt, end_pt) in self.coords:
                    indexes.append(self.coords.index((start_pt, end_pt)))
                if (end_pt, start_pt) in self.coords:
                    indexes.append(self.coords.index((end_pt, start_pt)))
                if any([type(edge) != type(intersection_edge) for i,edge in enumerate(self.edges) if i in indexes]):
                    print("TOOOOOOOOOOOODDDDDDDDDDOOOOOOOOOO")
                return 'pink'
            
            start_new_vtx = Vertex(*start_pt, self.all_vtx) if start_pt not in self.all_vtx else self.all_vtx[start_pt]
            end_new_vtx = Vertex(*end_pt, self.all_vtx) if end_pt not in self.all_vtx else self.all_vtx[end_pt] 
            
            if type(intersection_edge) != LineSegment:
                new_edge1 = Arc(vtx1=end_new_vtx, vtx2=start_new_vtx, clockwise=False or True, **arc_kwargs)
                new_edge2 = Arc(vtx1=start_new_vtx, vtx2=end_new_vtx, clockwise=True or False, **arc_kwargs)
            else:
                new_edge1 = LineSegment(vtx1=end_new_vtx, vtx2=start_new_vtx, **line_kwargs)
                new_edge2 = LineSegment(vtx1=start_new_vtx, vtx2=end_new_vtx, **line_kwargs)

            # check the type of the edge here to construct a new instance of it
            edge_start_before, edge_start_after = [[edge] for edge in start['edge'].split(start_new_vtx)] + ([] if not start['touches'] else [[]])
            edge_end_before, edge_end_after = [[edge] for edge in end['edge'].split(end_new_vtx)] + ([] if not end['touches'] else [[]])
            first_region_edge = new_edge1 if new_edge1.vtx1.xy == edge_end_before[0].vtx2.xy else new_edge2
            second_region_edge = new_edge1 if new_edge1.vtx1.xy == edge_start_before[0].vtx2.xy else new_edge2
            
            include_first_in_without_itx = (0 if start['touches'] and (start['i']+1==end['i'] or first_region_edge.vtx2.xy==self.edges[start['i']].vtx1.xy) else 1)
            # TODO MAKE THIS HANDLE LINESEGEMENTS ASWELL
            first_region_is_within_radius = pt_is_left_of_vector(*first_region_edge.center, *first_region_edge.vtx1.xy, *first_region_edge.vtx2.xy)
            
            reg_without_itx = OffsetRegion(
                edges = edge_start_after + self.edges[start['i']+include_first_in_without_itx:end['i']] + edge_end_before + [first_region_edge], 
                # checks = self.checks+[{**check, 'result': True}], all_regions = self.all_regions
                checks = self.checks+[{**check, 'result': first_region_is_within_radius}], all_regions = self.all_regions
            ) 
            # intersection
            exclude_first_in_itx = int(start['touches'] and self.edges[start['i']].vtx1.xy == first_region_edge.vtx2.xy)
            reg_itx = OffsetRegion(
                edges = edge_end_after + self.edges[end['i']+1:] +  self.edges[:start['i']] + (
                    edge_start_before if not (start['touches'] and self.edges[start['i']].vtx1.xy == first_region_edge.vtx2.xy) else []
                    ) + [second_region_edge], 
                # checks = self.checks+[{**check, 'result': True}], all_regions = self.all_regions
                checks = self.checks+[{**check, 'result': not first_region_is_within_radius}], all_regions = self.all_regions
            ) # Todo condition for when true
            
            if plot_split:
                fig, axs = plt.subplots(1,3, figsize=(12,3))
                for i in [0,1,2]:
                    ax = axs.flat[i]
                    coords = self.get_plot_coords()
                    x_coords = [x for x,y in coords]
                    y_coords = [y for x,y in coords]
                    x_min, x_max = min(x_coords), max(x_coords)
                    y_min, y_max = min(y_coords), max(y_coords)
                    x_dist, y_dist = x_max-x_min, y_max-y_min 
                    self.plot_single(
                        ax=ax, color='#ccc', plot_edges=False, x_lim=[x_min-x_dist/10, x_max+x_dist/10], y_lim=[y_min-y_dist/10, y_max+y_dist/10], plot_vertices=i==0, add_idx_edges=i==0,)
                    if i == 0:
                        first_region_edge.plot_single(ax=ax, linewidth=4)
                        ax.annotate(text='A', xy=first_region_edge.vtx1.xy)
                        ax.annotate(text='B', xy=first_region_edge.vtx2.xy)
                    if i == 1:
                        reg_without_itx.plot_single(ax=ax, color='green', alpha=0.5, hatch='/', plot_edges=False, x_lim=None, y_lim=None, plot_vertices=True)
                    if i == 2:
                        reg_itx.plot_single(ax=ax, color='red', alpha=0.5, plot_edges=False, x_lim=None, y_lim=None, plot_vertices=True)
                    GeoSeries([Point(intersection_edge.center).buffer(intersection_edge.r, 40)]).plot(ax=ax, edgecolor='black', facecolor='None')
                    if i == 0:
                        ax.set_title("T st:"+str(start['touches'])+", end:"+str(end['touches'])+ 'i st:'+str(start['i'])+' i end:'+str(end['i'])+
                                     ' nedges'+str(len(self.edges))+'+'+str(include_first_in_without_itx) + str(first_region_edge.vtx2.xy == self.edges[start['i']].vtx1.xy) )
                    elif i == 1:
                        ax.set_title("n vtcs:"+str(len(reg_without_itx.vertices)))
                    elif i == 2:
                        ax.set_title("n vtcs:"+str(len(reg_itx.vertices))+'ef'+str(exclude_first_in_itx))

            self.delete()

        # create arc
        return 'orange'
    #

    def calc_min_dist_to_pt(self, pt):
        """
        returns minimum distance of region to pt
        """
        return min([
            edge.calc_min_dist_to_pt(pt) for edge in self.edges
        ])
    #
    
    def calc_max_dist_to_pt(self, pt):
        """
        returns maximum distance of region to pt
        """
        return max([
            edge.calc_max_dist_to_pt(pt) for edge in self.edges
        ])
    #
    
    def transform_to_trgl(self, i:int):
        """
        Transform region from triangle into triangle i
        """
        if i == 1: return self 
        new_edges = []
        for edge in self.edges:
            new_edges.append(edge.transform_to_trgl(i=i))
        #
        
        if i % 2 == 1: 
            rotated_region = OffsetRegion(new_edges, checks=[], trgl_nr=i, all_regions=self.all_regions)
        else: 
            # reverse order for regions 2,4,6,8
            rotated_region = OffsetRegion(list(reversed(new_edges)), checks=[], trgl_nr=i, all_regions=self.all_regions)
        #
        
        rotated_region.contained_cells = tuple(sorted(
            [(lvl, (x,y)) for lvl, (x,y) in zip(
            [lvl for lvl,xy in self.contained_cells], 
            transform_cell_pattern([xy for lvl,xy in self.contained_cells], i))
        ]))
        
        rotated_region.overlapped_cells = tuple(sorted(
            [(lvl, (x,y)) for lvl, (x,y) in zip(
            [lvl for lvl,xy in self.overlapped_cells], 
            transform_cell_pattern([xy for lvl,xy in self.overlapped_cells], i))
        ]))
        
        return rotated_region
    #
    
    @staticmethod
    def merge_regions(regions:list, keep_old:bool=True, add_new:bool=False,):
        """
        Regions must are assumed to shared at least 1 edge, each point that is a combination of two points in regions to merged is assumed to lie within merge region 
        """

        # regions = list(all_regions.values())

        duplicated_edge_coords = set()
        all_edge_coords = set()
        all_edges = []
        for region in regions:
            for edge in region.edges:
                coords = edge.coords
                coords_rev = (coords[1], coords[0])
                
                if coords in all_edge_coords:
                    duplicated_edge_coords.update([coords, coords_rev])
                
                all_edge_coords.update([coords, coords_rev])
        
            all_edges.extend(region.edges)
            if not keep_old:
                region.delete()
        first_last_tpl_list = []
        edges_merged_region = []
        for region in regions:
            edges = region.edges
            duplicated_ids = [i for i,edge in enumerate(edges) if edge.coords in duplicated_edge_coords]
            non_duplicated_ids = [i for i,edge in enumerate(edges) if not edge.coords in duplicated_edge_coords]
            if len(non_duplicated_ids) == len(edges):
                raise ValueError("WEIRD°!")
            start = next((i for i in range(min(non_duplicated_ids), max(non_duplicated_ids)+1) if (i-1)%len(edges) not in non_duplicated_ids),None)
            end = next((i for i in range(min(non_duplicated_ids), max(non_duplicated_ids)+1) if (i+1)%len(edges) not in non_duplicated_ids), None)
            if start <= end:
                edges_merged_region.extend(edges[start:end+1])
            else:
                edges_merged_region.extend(edges[start:])
                edges_merged_region.extend(edges[:end+1])
            # check whether duplicate ids wrap around 
            
        merged_region = OffsetRegion(edges=edges_merged_region, checks=regions[0].checks, all_regions=region.all_regions, trgl_nr=min([reg.trgl_nr for reg in regions]))
        if not add_new:
            merged_region.delete()
        merged_region.contained_cells = regions[0].contained_cells
        merged_region.overlapped_cells = regions[0].overlapped_cells
        merged_region.nr = min([reg.nr for reg in regions])
        merged_region.trgl_nrs = [reg.trgl_nr for reg in regions]
        return merged_region
    #
#



"""
Tests for the private NumPy argument parsing functionality.
They mainly exists to ensure good test coverage without having to try the
weirder cases on actual numpy functions but test them in one place.

The test function is defined in C to be equivalent to (errors may not always
match exactly, and could be adjusted):

    def func(arg1, /, arg2, *, arg3):
        i = integer(arg1)  # reproducing the 'i' parsing in Python.
        return None
"""

import pytest

import numpy as np
from numpy._core._multiarray_tests import argparse_example_function as func


def test_invalid_integers():
    with pytest.raises(TypeError,
            match="integer argument expected, got float"):
        func(1.)
    with pytest.raises(OverflowError):
        func(2**100)


def test_missing_arguments():
    with pytest.raises(TypeError,
            match="missing required positional argument 0"):
        func()
    with pytest.raises(TypeError,
            match="missing required positional argument 0"):
        func(arg2=1, arg3=4)
    with pytest.raises(TypeError,
            match=r"missing required argument \'arg2\' \(pos 1\)"):
        func(1, arg3=5)


def test_too_many_positional():
    # the second argument is positional but can be passed as keyword.
    with pytest.raises(TypeError,
            match="takes from 2 to 3 positional arguments but 4 were given"):
        func(1, 2, 3, 4)


def test_multiple_values():
    with pytest.raises(TypeError,
            match=r"given by name \('arg2'\) and position \(position 1\)"):
        func(1, 2, arg2=3)


def test_string_fallbacks():
    # We can (currently?) use numpy strings to test the "slow" fallbacks
    # that should normally not be taken due to string interning.
    arg2 = np.str_("arg2")
    missing_arg = np.str_("missing_arg")
    func(1, **{arg2: 3})
    with pytest.raises(TypeError,
            match="got an unexpected keyword argument 'missing_arg'"):
        func(2, **{missing_arg: 3})


def test_too_many_arguments_method_forwarding():
    # Not directly related to the standard argument parsing, but we sometimes
    # forward methods to Python: arr.mean() calls np._core._methods._mean()
    # This adds code coverage for this `npy_forward_method`.
    arr = np.arange(3)
    args = range(1000)
    with pytest.raises(TypeError):
        arr.mean(*args)



from matplotlib.pyplot import (subplots as _plt_subplots)
from matplotlib.pyplot import get_cmap as _plt_get_cmap
from matplotlib.pyplot import close as _plt_close
from matplotlib.colors import LogNorm as _plt_LogNorm, Normalize as _plt_Normalize, LinearSegmentedColormap as _plt_LinearSegmentedColormap, ListedColormap as _plt_ListedColormap
from matplotlib.patches import Patch as _plt_Patch
from numpy import array as _np_array
from aabpl.illustrations.plot_utils import truncate_colormap, set_map_frame, plot_polygon
from shapely.geometry import Polygon as _shapely_Polygon, MultiPoint as _shapely_MultiPoint

def handle_plot_kwargs(default_kwargs:dict={}, **kwargs):
    plot_kwargs = {
        **default_kwargs
    }
    sub_group_kwargs = ['fig','ax','hlines','vlines','figsize','set_aspect','set_xticks','set_yticks','set_xlim','set_ylim','suptitle']
    # groups of kwargs from which only the last one shall be applied
    kwargs_groups = (('cmap','color','c'),)
    kwarg_to_group = {}
    for grp in kwargs_groups:
        for k in grp:
            if k not in kwarg_to_group:
                kwarg_to_group[k] = [x for x in grp if x!=k]
            else:
                kwarg_to_group[k].extend([x for x in grp if x!=k])
            #
        #
    #
    # 'hlines':{'color':'red', 'linewidth':1},
    # 'vlines':{'color':'red', 'linewidth':1},
    for kwarg, v in kwargs.items():
        if kwarg in kwargs_groups:
            for k in kwargs_groups[kwarg]:
                plot_kwargs.pop(k,None)
        plot_kwargs[kwarg] = v
    used_sub_groups = [k for k in sub_group_kwargs if k in plot_kwargs]
    def pop_all_subgroups(used_sub_groups=used_sub_groups, plot_kwargs=plot_kwargs):
        for k in used_sub_groups:
            plot_kwargs.pop(k,None)
    # plot_kwargs['pop_all'] = pop_all_subgroups
    return plot_kwargs


def create_plots_for_vars(
        grid,
        colnames:_np_array,
        filename:str="",
        save_kwargs:dict={},
        plot_kwargs:dict={}, 
        close_plot:bool=False,
):
    """
    TODO Descripiton
    """
    nrows = colnames.shape[0]
    ncols = 1 if len(colnames.shape)==1 else colnames.shape[1]
    # specify default plot kwargs and add defaults
    s = 1 if not 'figsize' in plot_kwargs else 0.1*plot_kwargs['figsize'][0]
    plot_kwargs = handle_plot_kwargs(default_kwargs={
        'fig': None,
        'axs': None,
        's': s,
        'cmap': 'Reds',
        'figsize': (10*ncols,8*nrows),
        'additional_varnames':[],
        'sample_area_colors': ["#ffffff",  "#bedbe6"]
    }, **plot_kwargs)
    save_kwargs = {'dpi':300, 'bbox_inches':"tight", **save_kwargs}
    
    plot_kwargs.pop('color', None)
    figsize = plot_kwargs.pop('figsize')
    fig = plot_kwargs.pop('fig')
    axs = plot_kwargs.pop('axs')
    
    sample_area_color, non_valid_area_color = plot_kwargs.pop('sample_area_colors')
    
    additional_varnames = plot_kwargs.pop('additional_varnames')
    if len(additional_varnames)>nrows:
        # TODO this is not compelted
        nrows = len(additional_varnames)
    if fig is None or axs is None:
        fig, axs = _plt_subplots(nrows,ncols, figsize=figsize)
    if not grid.sample_grid_bounds is None and not grid.sample_area is None:
        non_valid_area = _shapely_Polygon([
            (grid.sample_grid_bounds[0], grid.sample_grid_bounds[1]),
            (grid.sample_grid_bounds[2], grid.sample_grid_bounds[1]),
            (grid.sample_grid_bounds[2], grid.sample_grid_bounds[3]),
            (grid.sample_grid_bounds[0], grid.sample_grid_bounds[3])
            ]).difference(grid.sample_area) 
        xmin, xmax, ymin, ymax = grid.total_bounds.xmin, grid.total_bounds.xmax, grid.total_bounds.ymin, grid.total_bounds.ymax,
    else:    
        xmin, xmax, ymin, ymax = grid.total_bounds.xmin, grid.total_bounds.xmax, grid.total_bounds.ymin, grid.total_bounds.ymax,
    xs = grid.search.source.pts[grid.search.source.x]
    ys = grid.search.source.pts[grid.search.source.y]
    for i, colname in enumerate(colnames.flat):
        # SELECT AX (IF MULTIPLE)
        ax = axs.flat[i] if nrows > 1 else axs
        
        # SET TITLE
        ax_title = (colname)
        ax.set_title(ax_title)
        # CPOLOR NON SAMPLE AREA
        if not grid.sample_grid_bounds is None and not grid.sample_area is None:
            ax.set_facecolor(sample_area_color)
            ids_rndm_sample = grid.ids_rndm_sample
            col_min = int(round((grid.sample_grid_bounds[0]-grid.total_bounds.xmin)/grid.spacing,0))
            row_min = int(round((grid.sample_grid_bounds[1]-grid.total_bounds.ymin)/grid.spacing,0))
            col_max = int(round((grid.sample_grid_bounds[2]-grid.total_bounds.xmin)/grid.spacing-1,0))
            row_max = int(round((grid.sample_grid_bounds[3]-grid.total_bounds.ymin)/grid.spacing-1,0))
            # row_min, row_max = min([row for row,col in grid.ids_rndm_sample]), max([row for row,col in grid.ids_rndm_sample])
            # col_min, col_max = min([col for row,col in grid.ids_rndm_sample]), max([col for row,col in grid.ids_rndm_sample])
            X = _np_array([[not (row,col) in ids_rndm_sample for col in  range(col_min,col_max+1)] for row in range(row_min,row_max+1)[::-1]])
            cmap_binary = _plt_ListedColormap([sample_area_color, non_valid_area_color])
            extent = [grid.sample_grid_bounds[0],grid.sample_grid_bounds[2],grid.sample_grid_bounds[1],grid.sample_grid_bounds[3]]
            p = ax.imshow(X=X, interpolation='none', cmap=cmap_binary, extent=extent)#, To-Do the extent is imprecise as it does not cover the full grid only its points
            non_valid_patch = _plt_Patch(facecolor=non_valid_area_color, label='Non-valid area', edgecolor='black')
            sample_patch = _plt_Patch(facecolor=sample_area_color, label='Sample area', edgecolor='black')
            ax.legend(handles=[non_valid_patch, sample_patch], loc='best')
            plot_polygon(poly=non_valid_area, ax=ax, facecolor=non_valid_area_color, edgecolor='black')
        # ADD DISTRIUBTION PLOT
        c = grid.search.source.pts[colname]
        vmin=plot_kwargs['vmin'] if 'vmin' in plot_kwargs else c.min()
        vmax=plot_kwargs['vmax'] if 'vmax' in plot_kwargs else c.max(),
        # norm = plot_kwargs['norm'] if 'norm' in plot_kwargs else _plt_LogNorm(vmin=c.min(),vmax=c.max()) if (c.min() > 0) else _plt_LogNorm()
        norm = plot_kwargs['norm'] if 'norm' in plot_kwargs else _plt_LogNorm(vmin=c.min(),vmax=c.max()) if (c.min() > 0) else 'linear'
        plot_kwargs['cmap'] = truncate_colormap(_plt_get_cmap(plot_kwargs['cmap']),0.1,1)
        scttr = ax.scatter(x=xs,y=ys,c=c, norm=norm, **plot_kwargs)
        plot_polygon(poly=non_valid_area, ax=ax, facecolor="none", edgecolor='black')
        fig.colorbar(scttr, ax=ax)
        set_map_frame(ax=ax,xmin=xmin,xmax=xmax,ymin=ymin,ymax=ymax)

    if not fig is None:
        if filename:
            fig.savefig(filename, **save_kwargs)
        if close_plot:
            _plt_close(fig)
    return fig
    #
#


# replace imports with from imports
from numpy import (
    array as _np_array, 
    asarray as _np_asarray,
    arange as _np_arange, unique as _np_unique, linspace as _np_linspace, sign as _np_sign,
)
from numpy.linalg import norm as _np_linalg_norm
# from numpy.random import randint, random
# from matplotlib.animation import FuncAnimation, PillowWriter#
from matplotlib.patches import (Rectangle as _plt_Rectangle, Polygon as _plt_Polygon, Circle as _plt_Circle)
from matplotlib.colors import LinearSegmentedColormap as _plt_LinearSegmentedColormap
from matplotlib.colors import Normalize as _plt_Normalize
from matplotlib.axes._axes import Axes as _plt_Axes
from matplotlib.cm import ScalarMappable as _plt_ScalarMappable
from matplotlib.pyplot import subplots as _plt_subplots
from math import (
    log10 as _math_log10,
    sin as _math_sin,
    cos as _math_cos,
    asin as _math_asin,
    acos as _math_acos,
    atan2 as _math_atan2,
    pi as _math_pi)
from aabpl.utils.general import flatten_list, angle
from aabpl.utils.distances_to_cell import ( get_cell_farthest_vertex_to_point, get_cell_closest_point_to_points, )
from matplotlib.path import Path as _plt_Path
from matplotlib.patches import PathPatch as _plt_PathPatch
from matplotlib.collections import PatchCollection as _plt_PatchCollection
from shapely.geometry import (Polygon as _shapely_Polygon, MultiPolygon as _shapely_MultiPolygon)

# x = np.linspace(-6,6, num=100)
# y = np.linspace(-10,10, num=100)
# X,Y = np.meshgrid(x,y)
# Z = np.sin(X)/X+np.sin(Y)/Y

# fig, ax = plt.subplots()
# ax.contourf(X,Y,Z, alpha=.1)
# ax.contour(X,Y,Z, alpha=.4)

# class AnchoredHScaleBar(matplotlib.offsetbox.AnchoredOffsetbox):
#     """ size: length of bar in data units
#         extent : height of bar ends in axes units """
#     def __init__(self, size=1, extent = 0.03, label="", loc=2, ax=None,
#                  pad=0.4, borderpad=0.5, ppad = 0, sep=2, prop=None, 
#                  frameon=True, linekw={}, **kwargs):
#         if not ax:
#             ax = plt.gca()
#         trans = ax.get_xaxis_transform()
#         size_bar = matplotlib.offsetbox.AuxTransformBox(trans)
#         line = Line2D([0,size],[0,0], **linekw)
#         size_bar.add_patch(_plt_Rectangle(xy=(0,0), width=1, height=1, color='black'))
#         txt = matplotlib.offsetbox.TextArea(label, minimumdescent=False)
#         self.vpac = matplotlib.offsetbox.VPacker(children=[size_bar,txt],  
#                                  align="center", pad=ppad, sep=sep) 
#         matplotlib.offsetbox.AnchoredOffsetbox.__init__(self, loc, pad=pad, 
#                  borderpad=borderpad, child=self.vpac, prop=prop, frameon=frameon,
#                  **kwargs)

# ob = AnchoredHScaleBar(size=3, label="3 units", loc=4, frameon=True,
#                        pad=0.6,sep=4, linekw=dict(color="crimson"),) 
# ax.add_artist(ob)

def set_map_frame(ax, xmin:float, xmax:float, ymin:float, ymax:float):
    pad_x, pad_y = (xmax-xmin)/50, (ymax-ymin)/50
    ax.set_xlim([xmin-pad_x,xmax+pad_x])
    ax.set_ylim([ymin-pad_y,ymax+pad_y])
    xticks = [xmin, (xmin+xmax)/2, xmax]
    yticks = [ymin, (ymin+ymax)/2, ymax]
    ndigits = int(max([0, _math_log10(100/(xticks[1]-xticks[0])), _math_log10(100/(yticks[1]-yticks[0]))]))
    ax.set_xticks(xticks, labels=[int(round(t,ndigits)) if ndigits==0 else round(t,ndigits) for t in xticks])
    ax.set_yticks(yticks, labels=[int(round(t,ndigits)) if ndigits==0 else round(t,ndigits) for t in yticks])
    

# def _plt_colorbar(sc, extend='min', cax=add_color_bar_ax(fig,ax))

def add_color_bar_ax(fig,ax):
    if fig is None:
        cax = ax.inset_axes([ax.get_position().x1 + 0.01, ax.get_position().y0, 0.02 ,ax.get_position().height])
        return cax
    cax = fig.add_axes([ax.get_position().x1 + 0.01, ax.get_position().y0, 0.02 ,ax.get_position().height])
    return cax

def truncate_colormap(cmap, minval=0.0, maxval=1.0, n=200):
    new_cmap = _plt_LinearSegmentedColormap.from_list(
        'trunc({n},{a:.2f},{b:.2f})'.format(n=cmap.name, a=minval, b=maxval),
        cmap(_np_linspace(minval, maxval, n)))
    return new_cmap
#

def map_2D_to_rgb (
    x=None,
    y=None,
    xmin=None,
    xmax=None,
    ymin=None,
    ymax=None,
    rgb_xmin_ymin:tuple=(  2,253,102),
    rgb_xmax_ymin:tuple=(  2,  2,253),
    rgb_xmin_ymax:tuple=(253, 23,  2),
    skewness_exponent:float=0.5,

)->tuple:
    """
    if create_colorbars is True, it will not return rgb value but instead
    """

    if xmin is None: xmin = min(x)
    if xmax is None: xmax = max(x)
    if ymin is None: ymin = min(x)
    if ymax is None: ymax = max(x)
  
    x_share = max([1e-15, (x-xmin)/(xmax-xmin)])
    y_share = max([1e-15, (y-ymin)/(ymax-ymin)])
    m_share = ((x_share+y_share)/2)**skewness_exponent
    x_ratio = x_share/(x_share+y_share)
    y_ratio = y_share/(x_share+y_share)
    rgb = tuple([
        ((1-m_share) * v_0 + m_share * (x_ratio * v_x + y_ratio * v_y))/255
        for v_0, v_x, v_y in zip(rgb_xmin_ymin, rgb_xmax_ymin, rgb_xmin_ymax)
        ])
    return rgb

def get_2D_rgb_colobar_kwargs(
    xmin,
    xmax,
    ymin,
    ymax,
    rgb_xmin_ymin:tuple=(  2,253,102),
    rgb_xmax_ymin:tuple=(  2,  2,253),
    rgb_xmin_ymax:tuple=(253, 23,  2),
    skewness_exponent:float=0.5,
):
    kwargs = locals()
    steps = 20
    return ({'mappable': _plt_ScalarMappable(
                norm = _plt_Normalize(vmin=xmin, vmax=xmax, clip=False),
                cmap = _plt_LinearSegmentedColormap.from_list('', [map_2D_to_rgb(x=v, y=ymin, **kwargs) for v in _np_linspace(xmin, xmax, steps)])
                )
            },
            {'mappable': _plt_ScalarMappable(
                norm = _plt_Normalize(vmin=ymin, vmax=ymax, clip=False),
                cmap = _plt_LinearSegmentedColormap.from_list('', [map_2D_to_rgb(x=xmin, y=v, **kwargs) for v in _np_linspace(ymin, ymax, steps)])
                )
            },
            {'mappable': _plt_ScalarMappable(
                norm = _plt_Normalize(vmin=0, vmax=1, clip=False),
                cmap = _plt_LinearSegmentedColormap.from_list('', [map_2D_to_rgb(x=x, y=y, **kwargs) for x,y in zip(_np_linspace(xmin, xmax, steps), _np_linspace(ymin, ymax, steps))])
                )
            }
            )
    

#################### RECTANGLES ########################################################

def create_grid_cell_patches(
        grid_spacing,
        ax_min,
        ax_max,
        contain_cells_row_col:_np_array,
        overlap_cells_row_col:_np_array,
        contain_triangle_cells_row_col:_np_array=_np_array([])
    )->list:
    """
    Return list of grid cell patches
    """
    gridCellPatches = []

    # choose cell_steps_max s.t. the grid fills the complete plot area
    cell_steps_max = int(ax_max/grid_spacing+2)
    for j in range(cell_steps_max):
        for k in range(cell_steps_max):
            rect_color = (
                'yellow' if ((j,k) in contain_triangle_cells_row_col) else
                'green' if ((j,k) in contain_cells_row_col or (k,j) in contain_cells_row_col) else 
                'red' if ((j,k) in overlap_cells_row_col or (k,j) in overlap_cells_row_col) else 
                'grey'
            )
            # rect_edgecolor = 'red' if j==0 and k==0 else '#000'
            for pj in [-1,1]:
                for pk in [-1,1]:
                    append_rect = False
                    if (-1 in [pj,pk]) and (j!=0) and (k!=0):
                        # add to plot if top right is within plot area
                        if (pj*j+.5)*grid_spacing>ax_min and (pk*k+.5)*grid_spacing>ax_min:
                            append_rect = True
                    else:
                        # append all rect in top right quarter 
                        append_rect = True
                    
                    if append_rect:
                        gridCellPatches.append(_plt_Rectangle(
                            ((j*pj-.5)*grid_spacing, (k*pk-.5)*grid_spacing), grid_spacing, grid_spacing, 
                            linewidth=.7, facecolor=rect_color, edgecolor='#444', alpha=.5
                            ))
    return gridCellPatches
#

def create_grid_cell_patches_by_type(
        grid_spacing,
        contain_cells_row_col:_np_array,
        overlap_cells_row_col:_np_array,
        contain_triangle_cells_row_col:_np_array=_np_array([]),
        outside_cells_row_col:_np_array=_np_array([])
    )->list:
    """
    Return list of grid cell patches
    """
    gridCellPatches = []

    for cell_by_type, color_for_type in zip(
        [contain_cells_row_col, contain_triangle_cells_row_col, overlap_cells_row_col, outside_cells_row_col],
        ['green', 'yellow', 'red', 'grey']
    ):
        for cell in cell_by_type:
            gridCellPatches.append(_plt_Rectangle(
                ((cell[1]-.5)*grid_spacing, (cell[0]-.5)*grid_spacing), grid_spacing, grid_spacing, 
                linewidth=.7, facecolor=color_for_type, edgecolor='#444', alpha=.3
                ))
    return gridCellPatches
#

def create_grid_cell_rectangles(
    cells:_np_array,
    grid_spacing,
    facecolor:str='green',
    edgecolor='#444',
    alpha=.3,
    x_off:float=0,
    y_off:float=0,
)->list:
    """
    Return list of grid cell patches
    """
    gridCellPatches = []
    for cell in cells:
        gridCellPatches.append(_plt_Rectangle(
            xy = (
                (cell[1] - .5) * grid_spacing + x_off,
                (cell[0] - .5) * grid_spacing + y_off
            ), width=grid_spacing, height=grid_spacing, 
            linewidth=.7, facecolor=facecolor, edgecolor=edgecolor, alpha=alpha
            ))
    return gridCellPatches
#

def add_grid_cell_rectangles_by_color(
    list_of_cells:list,
    list_of_facecolors:list,
    ax:_plt_Axes,
    grid_spacing:float=1,
    x_off:float=0,
    y_off:float=0,
): 
    # create patches
    for cells, facecolor in zip(list_of_cells, list_of_facecolors):
        cell_patches = create_grid_cell_rectangles(
            cells=cells,
            grid_spacing=grid_spacing, 
            facecolor=facecolor,
            x_off=x_off,
            y_off=y_off,
        )
        for cell_patch in cell_patches:
            ax.add_patch(cell_patch)
#

def create_trgl1_patch(
        side_length:float = 1,
        facecolor:str='None', 
        edgecolor:str='green',
        x_off:float=0,
        y_off:float=0,
        **kwargs,
) -> list:
    """
    
    """
    poly_coords = (
        # top right
        [(0+x_off,0+y_off), (side_length+x_off, 0+y_off), (side_length+x_off, side_length+y_off)]
    )

    return _plt_Polygon(poly_coords, facecolor=facecolor, edgecolor=edgecolor, **kwargs)
#


#################### CIRCLES ########################################################

def create_circle_arc_coords(
    pts:_np_array, 
    arc:_np_array,
    r:float,
    rotation_angles:float=None
):
    if rotation_angles is None:
        n=len(pts)
        rotation_angles=_np_arange(n) / n * 2 *_math_pi
    r=r
    poly_coords = flatten_list([[(r*_math_cos(t) + pt_x, r*_math_sin(t) + pt_y) for t in angle+arc]  for (pt_x,pt_y),angle in zip(pts, rotation_angles)])

    return poly_coords

def create_buffered_square_patch(
        side_length:float,
        r:float=750,
        nsteps:int = 25,
        facecolor:str='None', 
        edgecolor:str='green',
        x_off:float=0,
        y_off:float=0,
        **kwargs,
) -> _plt_Polygon:
    """
    
    """
    x = side_length / 2
    poly_coords = create_circle_arc_coords(pts=(
            (+x + x_off, +x + y_off), # top right
            (-x + x_off, +x + y_off), # top left
            (-x + x_off, -x + y_off), # bottom left
            (+x + x_off, -x + y_off), # bottom right
        ),
        arc = _math_pi * _np_linspace(0, .5,nsteps),
        r = r,
    )

    return _plt_Polygon(poly_coords, facecolor=facecolor, edgecolor=edgecolor, **kwargs)
#
  
def create_debuffered_square_patch(
        side_length:float,
        r:float=750,
        nsteps:int = 25,
        facecolor:str='None', 
        edgecolor:str='red',
        x_off:float=0,
        y_off:float=0,
        **kwargs,
) -> _plt_Polygon:
    """
    
    """
    x = side_length/2
    alpha = _math_acos(x/2/r)/_math_pi
    beta = 0.5-alpha
    print("ALPA",alpha)
    poly_coords = create_circle_arc_coords(pts=(
            (-x + x_off, -x + y_off), # top right
            (+x + x_off, -x + y_off), # top left
            (+x + x_off, +x + y_off), # bottom left
            (-x + x_off, +x + y_off), # bottom right
        ),
        arc=_math_pi * _np_linspace(beta, alpha,nsteps,endpoint=True),
        r=r,
    )
    return _plt_Polygon(poly_coords,facecolor=facecolor,edgecolor=edgecolor, **kwargs)
#

def create_buffered_trgl1_patch(
        side_length:float = 250,
        r:float = 750,
        nsteps:int = 25,
        facecolor:str='None', 
        edgecolor:str='yellow',
        endpoint=True,
        x_off:float=0,
        y_off:float=0,
        **kwargs,
) -> list:
    """
    
    """
    x = side_length
    poly_coords =  (
        # top right/left
        [(r*_math_cos(t) + x + x_off, r*_math_sin(t) + x + y_off) for t in _math_pi * _np_linspace(0, 0.75,nsteps)] +
        # top/bottom left 
        [(r*_math_cos(t) + x_off, r*_math_sin(t) + y_off) for t in _math_pi * _np_linspace(0.75, 1.5,nsteps)] +
        # bottom right
        [(r*_math_cos(t) + x + x_off, r*_math_sin(t) + y_off) for t in _math_pi * _np_linspace(1.5, 2,nsteps)]
    )
    return _plt_Polygon(poly_coords, facecolor=facecolor, edgecolor=edgecolor, **kwargs)
#

def create_debuffered_trgl1_patch(
        side_length:float = 250,
        r:float = 750,
        nsteps:int = 25,
        x_off:float=0,
        y_off:float=0,
        facecolor:str='None', 
        edgecolor:str='black',
        **kwargs,
) -> list:
    """
    
    """
    x = side_length
    alpha = _math_acos(x/2/r)/_math_pi
    beta = 0.5-alpha
    poly_coords = (
        # bottom/top right
        [(r*_math_cos(t) + x_off, r*_math_sin(t) + y_off) for t in _math_pi * _np_linspace(-0.25, alpha,nsteps,endpoint=True)] +
        # top left
        [(r*_math_cos(t) + x + x_off, r*_math_sin(t) + y_off) for t in _math_pi * _np_linspace(0.5+beta, 0.5+alpha,nsteps,endpoint=True)] +
        # bottom left/right 
        [(r*_math_cos(t) + x + x_off, r*_math_sin(t) + x + y_off) for t in _math_pi * _np_linspace(1+beta, 1.75,nsteps,endpoint=True)]
    )

    return _plt_Polygon(poly_coords, facecolor=facecolor, edgecolor=edgecolor, **kwargs)

def  dual_circle_union_patch(
        centroids:_np_array,
        r:float,
        nsteps:int=100,
        **kwargs,
        ):
    dist = _np_linalg_norm(centroids[1]-centroids[0])
    if dist >= r:
        print("too far apart. 2 circles are not implemented")
        return
    alpha = _math_acos(dist/2/r)
    left_x, left_y = centroids[0]
    right_x, right_y = centroids[1]
    slope_angle = angle(left_x, left_y, right_x, right_y)

    poly_coords = (
        # left half
        [(r*_math_cos(t) + left_x, r*_math_sin(t) + left_y) for t in -slope_angle + _np_linspace(alpha, 2*_math_pi-alpha,nsteps)] +
        # right half
        [(r*_math_cos(t) +right_x, r*_math_sin(t) + right_y) for t in -slope_angle + _np_linspace(_math_pi+alpha, 3*_math_pi-alpha,nsteps)]
    )
    
    return _plt_Polygon(poly_coords, **kwargs)
#

def add_circle_patches(
    ax:_plt_Axes,
    list_of_cells:list,
    list_of_edgecolors:list,
    list_of_tuples_check_farthest_closest:list,
    edgecolor_outside_center_cell=True,
    convex_set_boundaries:_np_array= _np_array([(-0.5,-0.5), (0.5,-0.5),(0.5,0.5),(-0.5,0.5)]),
    grid_spacing:float=1, 
    r:float=3, 
    **kwargs,
):
    circles_outside_center_cell = []
    circles_inside_center_cell = []
    for (cells, edgecolor, (check_farthest, check_closest)) in zip(list_of_cells, list_of_edgecolors, list_of_tuples_check_farthest_closest):
        if (
            (type(edgecolor_outside_center_cell)==bool and edgecolor_outside_center_cell == True) or 
            (type(edgecolor_outside_center_cell)==str and not edgecolor_outside_center_cell=='None')
            ):
            edgecolor_outside = edgecolor if (edgecolor_outside_center_cell == True) else edgecolor_outside_center_cell
        
        for cell in cells:
            (row,col) = cell
            farthest_points = _np_unique(
                    get_cell_farthest_vertex_to_point(
                        convex_set_boundaries,
                        cell
                ), axis=0)*grid_spacing if check_farthest else []
            if check_closest:
                if row == 0 or col == 0:
                    # 
                    if edgecolor_outside not in [False, None, 'None']:
                        circles_outside_center_cell.append(create_buffered_square_patch(
                            side_length=grid_spacing, r=r,
                            edgecolor=edgecolor_outside,
                            x_off = (col - _np_sign(col))*grid_spacing,
                            y_off = (row - _np_sign(row))*grid_spacing,
                            **kwargs
                            ))
                    # circle_patches_segment.append()
                    pass
                else:
                    closest_cell_vertex = get_cell_closest_point_to_points(
                        convex_set_boundaries,
                        cell
                    )
                    if edgecolor_outside not in [False, None, 'None']:
                        for xy in _np_unique(closest_cell_vertex,axis=0):
                            circles_outside_center_cell.append(_plt_Circle(
                                xy=xy*grid_spacing, r=r, facecolor='None',edgecolor=edgecolor_outside, **kwargs
                            ))
                    # circle_patches_segment.append()
                    pass
            if len(farthest_points)==2:
                if row != 0 or col != 0:  
                    circles_outside_center_cell.append(
                        dual_circle_union_patch(farthest_points, r=r, facecolor='None',edgecolor=edgecolor_outside, **kwargs)
                    )
                else:
                    print("WARNING THIS IS NOT IMPLEMENTED",farthest_points)
                    for farthest_point in farthest_points:
                        circles_outside_center_cell.append(_plt_Circle(
                            xy=farthest_point, r=r, facecolor='None',edgecolor=edgecolor_outside, **kwargs
                        ))        
            else:
                if len(farthest_points)>2:
                    print("WARNING THIS IS NOT IMPLEMENTED", farthest_points)
                for farthest_point in farthest_points:
                    circles_outside_center_cell.append(_plt_Circle(
                        xy=farthest_point, r=r, facecolor='None',edgecolor=edgecolor_outside, **kwargs
                    ))        
                    # circle_patches_segment.append()
            #
        # 
    # 
                
    # circles_inside_center_cell.append(...)
    for patch_to_add in circles_outside_center_cell+circles_inside_center_cell:
        ax.add_patch(patch_to_add)
#

def create_circle_patches(
        grid_spacing:float,
        r:float=750,
        x_off:float=0,
        y_off:float=0,
        nsteps:int = 25,
        facecolor:str='None', 
        edgecolor:str='green',
        linewidth:float=3,
) -> list:
    """
    
    """
    outer_poly_patch = create_buffered_square_patch(
        side_length=grid_spacing,
        r=r,
        nsteps=nsteps,
        facecolor=facecolor, 
        edgecolor='green',
        linewidth=3,
        x_off=x_off,
        y_off=y_off,
    )
    inner_poly_patch = create_debuffered_square_patch(
        side_length=grid_spacing,
        r=r,
        nsteps=nsteps,
        facecolor=facecolor, 
        edgecolor='green',
        linewidth=3,
        x_off=x_off,
        y_off=y_off,
    )
    return [outer_poly_patch, inner_poly_patch]
#

def plot_polygon(poly, ax=None, **kwargs):
    if ax is None:
        fig, ax = _plt_subplots()

    geoms = [poly] if type(poly) == _shapely_Polygon else list(poly.geoms)
    for geom in geoms:
        path = _plt_Path.make_compound_path(
            _plt_Path(_np_asarray(geom.exterior.coords)[:, :2]),
            *[_plt_Path(_np_asarray(ring.coords)[:, :2]) for ring in geom.interiors]
            )

        patch = _plt_PathPatch(path, **kwargs)
        collection = _plt_PatchCollection([patch], **kwargs)
        
        ax.add_collection(collection, autolim=True)
        ax.autoscale_view()
    return collection
#


from numpy import array as _np_array
from pandas import (DataFrame as _pd_DataFrame, cut as _pd_cut, concat as _pd_concat) 
from aabpl.utils.general import ( DataFrameRelation, arr_to_tpls)
from aabpl.utils.distances_to_cell import get_cells_relevant_for_disk_by_type
from .two_dimensional_weak_ordering_class import gen_weak_order_rel_to_convex_set
from .pts_to_cells import assign_points_to_cells, aggregate_point_data_to_cells
from .pts_to_offset_regions import assign_points_to_mirco_regions
from .pts_radius_search import aggregate_point_data_to_disks_vectorized
from aabpl.testing.test_performance import time_func_perf


################ DiskSearchSource ######################################################################################
class DiskSearchObject(object):

    def assign_pts_to_cells(
        self,
        silent:bool = False
    ):
        return assign_points_to_cells(
            grid=self.grid,
            pts=self.pts,
            y=self.y,
            x=self.x,
            row_name=self.row_name,
            col_name=self.col_name,
            silent=silent,
        )
    #
#

################ DiskSearchSource ######################################################################################
class DiskSearchSource(DiskSearchObject):
    def __init__(
        self,
        grid,
        pts:_pd_DataFrame,
        c:list=[],
        y:str='lat',
        x:str='lon',
        row_name:str='id_y',
        col_name:str='id_x',
        sum_suffix:str='_750m',
    ):
        self.grid = grid 
        self.pts = pts
        self.y = y
        self.x = x
        self.row_name = row_name
        self.col_name = col_name
        self.cell_region_name = next(('cell_reg'+str(i) for i in (['']+list(range(len(pts.columns)))) if not 'cell_reg'+str(i) in pts.columns)) 
        self.sum_suffix = sum_suffix
        self.aggregate_columns = [str(column)+sum_suffix for column in c]

    #next(('helper_col'+i for i in (['']+list(range(len(pts_target.columns))))))

    def assign_pts_to_cell_regions(
            self,
            plot_cell_reg_assign:dict=None,
            plot_offset_checks:dict=None,
            plot_offset_regions:dict=None,
            plot_offset_raster:dict=None,
            silent:bool=False,
    ):
        # TODO 
        return assign_points_to_mirco_regions( 
        # return assign_points_to_cell_regions(
            grid=self.grid,
            pts=self.pts,
            r=self.grid.search.r,
            nest_depth=self.grid.nest_depth,
            include_boundary=self.grid.search.include_boundary,
            y=self.y,
            x=self.x,
            row_name=self.row_name,
            col_name=self.col_name,
            cell_region_name=self.cell_region_name,
            plot_cell_reg_assign=plot_cell_reg_assign,
            plot_offset_checks=plot_offset_checks,
            plot_offset_regions=plot_offset_regions,
            plot_offset_raster=plot_offset_raster,
            silent=silent,
        )
    #
#

################ DiskSearchTarget ######################################################################################
class DiskSearchTarget(DiskSearchObject):
    def __init__(
        self,
        grid,
        pts:_pd_DataFrame,
        c:list=[],
        y:str='lat',
        x:str='lon',
        row_name:str='id_y',
        col_name:str='id_x',
    ):
        self.grid = grid 
        self.pts = pts
        self.c = c
        self.x = x
        self.y = y
        self.row_name = row_name
        self.col_name = col_name
        
        # prepare dicts for fast lookup of values for point ids
        self.pt_id_to_xy_coords = {
            pt_id:xy for (pt_id,xy) in zip(pts.index, pts[[x,y]].values)
        }
        self.pt_id_to_vals = {
            pt_id:pt_vals for (pt_id,pt_vals) in zip(pts.index, pts[c].values)
        }
    #
    
    def aggregate_pt_data_to_cells(
            self,
            silent
    ):
        return aggregate_point_data_to_cells(
            grid=self.grid,
            pts=self.pts,
            y=self.y,
            x=self.x,
            c=self.c,
            row_name=self.row_name,
            col_name=self.col_name,
            nest_depth=self.grid.nest_depth,
            silent=silent,
        )
    #
#
   

################ DiskSearch ######################################################################################
class DiskSearch(object):
    def __init__(
        self,
        grid,
        r:float=0.0075,
        nest_depth:int=None,
        exclude_pt_itself:bool=True,
        weight_valid_area:str=None,
        include_boundary:bool=False,
    ):
            
        """
        
        """
        # link to grid
        grid.search = self
        self.grid = grid
        self.exclude_pt_itself = exclude_pt_itself
        self.weight_valid_area = weight_valid_area
        self.include_boundary = include_boundary

        self.update_search_params(
            grid=grid,
            exclude_pt_itself=exclude_pt_itself,
            weight_valid_area=weight_valid_area,
            r=r,
            nest_depth=nest_depth,
            include_boundary=include_boundary,
        )
        #
    #

    @time_func_perf
    def update_search_params(
        self,
        grid,
        exclude_pt_itself:bool=None,
        weight_valid_area:str=None,
        r:float=None,
        nest_depth:int=None,
        include_boundary:bool=None,
        relation_tgt_to_src:str=None,
    ):
        if exclude_pt_itself is not None:
            self.exclude_pt_itself = exclude_pt_itself
        if weight_valid_area is not None:
            self.weight_valid_area = weight_valid_area
        if relation_tgt_to_src is not None:
            self.relation_tgt_to_src = relation_tgt_to_src
        
        # early return if r and include_boundary have not changed
        if hasattr(self, 'r') and hasattr(self, 'include_boundary'):
            if self.r == r and self.include_boundary == include_boundary:
                return
        
        # store params
        self.r = r
        self.nest_depth = nest_depth if not nest_depth is None else grid.nest_depth
        self.include_boundary = include_boundary
        self.overlap_checks = []
        self.contain_checks = []
        
        # get relative position of cells that are always included within r for current gridsize    
        (
        self.cells_contained_in_all_disks, 
        self.cells_contained_in_all_trgl_disks, 
        self.cells_maybe_overlapping_a_disk, 
        self.cells_maybe_overlapping_a_trgl_disk
        ) = get_cells_relevant_for_disk_by_type(
                grid_spacing=grid.spacing,
                r=r,
                include_boundary=include_boundary,
        )
        # hierarchically order all cells with respect to any point in triangle. 
        # Some cells are at least as far away as others 
        # e.g. (2,2) is weakly closer than (-2,-2) as any pt in triangle is P(x>=0,y>=0)
        # e.g. (2,3) is weakly closer than (?,?) as any pt in triangle is P(x,y<=0.5*grid.spacing)
        triangle_1_vertices = _np_array([[0,0],[0.5,0],[0.5,0.5]])
        vertices_is_inside_triangle_1 = _np_array([True,True,False],dtype=bool)
        # TODO r,grid_spacing, include_boundary could be removed from weak_order_tree generation
        self.weak_order_tree = gen_weak_order_rel_to_convex_set(
                cells=self.cells_maybe_overlapping_a_trgl_disk,
                convex_set_vertices = triangle_1_vertices,
                vertex_is_inside_convex_set = vertices_is_inside_triangle_1,
                r=r,
                grid_spacing=grid.spacing,
                include_boundary=include_boundary,
        )
        # TODO lvl is hard coded to 0. IF MULTIPLE LEVELS ARE CONSIDERED THIS MIGHT NEED TO BE ADJUSTED in get_cells_relevant_for_disk_by_type
        self.cells_contained_in_all_disks = [(0,(row,col)) for (row,col) in arr_to_tpls(self.cells_contained_in_all_disks,int)]
    #
    
    # @time_func_perf
    def check_if_tgt_df_contains_src_df(
        self,
        silent:bool=False,
    )->bool:
        if not hasattr(self, 'target'): return False
        if not hasattr(self, 'source'): return False
        if not hasattr(self.target, 'pts'): return False
        if not hasattr(self.source, 'pts'): return False
        return DataFrameRelation.check_if_df_is_contained(self.source.pts, self.target.pts,silent=silent)
    
    # @time_func_perf
    def check_if_search_obj_already_exist(
        self,
        pts:_pd_DataFrame,
        obj:str=['source','target'],
        silent:bool=False,
        **kwargs
    ):
        """
        check if search sortarget already created
        kwarg at pos 0 is pandas.DataFrame and will thus be checked for
        equlity by .equals insted of ==  
        Args:
        self : DiskSearch
          Checks on disk searchs object
        Returns:
        
        """
        
        # check if this attribute is already set as source
        alr_added_pts_to_grid = (
            hasattr(self, 'target') and 
            all([hasattr(self.target, k) and 
                v == getattr(self.target, k)
                for k,v in kwargs.items()]) and
                hasattr(self.target, 'pts') and 
                DataFrameRelation.check_if_df_is_contained(pts, self.target.pts,silent=silent)
        )
        alr_assg_to_cell_regions = obj=='source' and (
            hasattr(self, 'source') and all([
                hasattr(self.source, k) and 
                v == getattr(self.source, k)
                for k,v in kwargs.items()]) and
                hasattr(self.source, 'pts') and 
                DataFrameRelation.check_if_df_is_contained(pts, self.source.pts,silent=silent)
        )
        alr_assg_to_cells = (alr_added_pts_to_grid or alr_assg_to_cell_regions)
        
        return (alr_assg_to_cells, alr_assg_to_cell_regions, alr_added_pts_to_grid)
    #

    @time_func_perf
    def set_source(
        self,
        pts:_pd_DataFrame,
        c:list=[],
        y:str='lat',
        x:str='lon',
        row_name:str='id_y',
        col_name:str='id_x',
        sum_suffix:str='_750m',
        plot_cell_reg_assign:dict=None,
        plot_offset_checks:dict=None,
        plot_offset_regions:dict=None,
        plot_offset_raster:dict=None,
        silent:bool=False,
    ):
        """
        TODO also make shortcut if  grid.search.tgt_df_contains_src_df
        """
        (alr_assg_to_cells,
        alr_assg_to_cell_regions,
        alr_added_pts_to_grid) = self.check_if_search_obj_already_exist(
            **dict(list(locals().items())[1:8])
        )

        self.source = DiskSearchSource(
            grid=self.grid,
            pts=pts,
            c=c,
            y=y,
            x=x,
            row_name=row_name,
            col_name=col_name,
            sum_suffix=sum_suffix,
        )
        
        self.tgt_df_contains_src_df = self.check_if_tgt_df_contains_src_df(silent=silent)

        if not alr_assg_to_cells:
            self.source.assign_pts_to_cells(silent=silent,)
            # self.source.pts.sort_values(
            #     [self.source.row_name, self.source.col_name],
            #     inplace=True
            # )
        #
        if not alr_assg_to_cell_regions:
            self.source.assign_pts_to_cell_regions(
                plot_cell_reg_assign=plot_cell_reg_assign,
                plot_offset_checks=plot_offset_checks,
                plot_offset_regions=plot_offset_regions,
                plot_offset_raster=plot_offset_raster,
                silent=silent,
            )
            # self.source.pts.sort_values(
            #     [self.source.row_name, self.source.col_name, self.source.cell_region_name],
            #     inplace=True
            # )
        #
    #

    @time_func_perf
    def set_target(
        self,

        pts:_pd_DataFrame,
        c:list=['employment'],
        y:str='lat',
        x:str='lon',
        row_name:str='id_y',
        col_name:str='id_x',

        silent:bool=False,
    ):
        
        (alr_assg_to_cells,
        alr_assg_to_cell_regions,
        alr_added_pts_to_grid) = self.check_if_search_obj_already_exist(
            **dict(list(locals().items())[1:8])
        )

        self.target = DiskSearchTarget(
            grid=self.grid,
            pts=pts,
            c=c,
            y=y,
            x=x,
            row_name=row_name,
            col_name=col_name,
        )

        self.tgt_df_contains_src_df = self.check_if_tgt_df_contains_src_df(silent=silent)

        if not alr_assg_to_cells:
            self.target.assign_pts_to_cells(silent=silent,)

            # also sort according to grid cell region!
            # self.target.pts.sort_values(
            #     [self.target.row_name, self.target.col_name],
            #     inplace=True
            # )
        #
        if not alr_added_pts_to_grid:
            self.target.aggregate_pt_data_to_cells(silent=silent,)
        #
    #

    def perform_search(
            self,
            plot_pt_disk:dict=None,
            silent:bool=False,
    ):
        
        return aggregate_point_data_to_disks_vectorized(
            grid=self.grid,
            pts_source=self.source.pts,
            pts_target=self.target.pts,
            r=self.r,
            c=self.target.c,
            y=self.source.y,
            x=self.source.x,
            row_name=self.source.row_name,
            col_name=self.source.col_name,
            cell_region_name=self.source.cell_region_name,
            sum_suffix=self.source.sum_suffix,
            exclude_pt_itself=self.exclude_pt_itself,
            weight_valid_area=self.weight_valid_area,
            plot_pt_disk=plot_pt_disk,
            silent=silent,
        )
    #
#
    
    
